"""Preprocessing pipeline for Finding_and_Measuring_Lungs_in_CT_Data dataset."""
import os
from dataclasses import asdict, dataclass, field
from typing import Any

from base.extractors import BaseImgIdExtractor, BaseStudyIdExtractor, study_id
from base.pipeline import BasePipeline, PipelineArgs
from config.dataset_config import DatasetArgs, finding_and_measuring_lungs
from constants import IMG_FOLDER_NAME, MASK_FOLDER_NAME
from steps import (
    AddUmieIds,
    ConvertTif2Png,
    CopyMasks,
    CreateFileTree,
    DeleteTempFiles,
    GetFilePaths,
    RecolorMasks,
    StoreSourcePaths,
)


class ImgIdExtractor(BaseImgIdExtractor):
    """Extractor for image IDs specific to the Finding_and_Measuring_Lungs_in_CT_Data dataset."""

    def _extract(self, img_path: str) -> str:
        """Retrieve image id from path."""
        return "0.png"


class StudyIdExtractor(BaseStudyIdExtractor):
    """Extractor for study IDs specific to the Finding_and_Measuring_Lungs_in_CT_Data dataset."""

    def _extract(self, img_path: str) -> str:
        """Get study ID for dataset."""
        # Getting study id depends on location of the file.
        # Study_id is retrieved in a different way when image already is moved to target directory with new name.
        return os.path.basename(img_path).split("_")[-3]


@dataclass
class FindingAndMeasuringLungsPipeline(BasePipeline):
    """Preprocessing pipeline for finding_and_measuring_lungs dataset."""

    name: str = "finding_and_measuring_lungs"  # dataset name used in configs
    steps: tuple = (
        ("create_file_tree", CreateFileTree),
        ("get_file_paths", GetFilePaths),
        ("store_source_paths", StoreSourcePaths),
        ("convert_tif2png", ConvertTif2Png),
        ("copy_png_masks", CopyMasks),
        ("recolor_masks", RecolorMasks),
        ("add_new_ids", AddUmieIds),
        ("delete_temp_files", DeleteTempFiles),
    )
    dataset_args: DatasetArgs = field(default_factory=lambda: finding_and_measuring_lungs)
    pipeline_args: PipelineArgs = field(
        default_factory=lambda: PipelineArgs(
            image_folder_name=IMG_FOLDER_NAME,
            mask_folder_name=MASK_FOLDER_NAME,
            img_prefix="images",  # prefix of the source image file names
            segmentation_prefix="masks",  # prefix of the source mask file names
            mask_selector="2d_masks",
            study_id_extractor=StudyIdExtractor(),
            img_id_extractor=ImgIdExtractor(),
        )
    )

    def prepare_pipeline(self) -> None:
        """Post initialization actions."""
        # Add dataset specific arguments to the pipeline arguments
        self.args: dict[str, Any] = dict(**self.args, **asdict(self.pipeline_args))



"""
This module contains the pipeline classes for opensource medical imaging datasets supported by UMIE.

Each pipeline class defines how to preprocess a specific dataset.
A pipline class consists of a set of steps that are executed in a specific order.
The pipelines included in this module are:
- AlzheimersPipeline
- BrainMETSharePipeline
- BrainTumorClassificationPipeline
- BrainTumorDetectionPipeline
- BrainTumorProgressionPipeline
- BrainWithIntracranialHemorrhagePipeline
- ChestXray14Pipeline
- COCAPipeline
- CoronaHackPipeline
- COVID19DetectionPipeline
- FindingAndMeasuringLungsPipeline
- KITS23Pipeline
- KneeOsteoarthritisPipeline
- LITSPipeline
"""
from .alzheimers import AlzheimersPipeline
from .brain_met_share import BrainMETSharePipeline
from .brain_tumor_classification import BrainTumorClassificationPipeline
from .brain_tumor_detection import BrainTumorDetectionPipeline
from .brain_tumor_progression import BrainTumorProgressionPipeline
from .brain_with_intracranial_hemorrhage import BrainWithIntracranialHemorrhagePipeline
from .chest_xray14 import ChestXray14Pipeline
from .coca import COCAPipeline
from .coronahack import CoronaHackPipeline
from .covid19_detection import COVID19DetectionPipeline
from .finding_and_measuring_lungs import FindingAndMeasuringLungsPipeline
from .kits23 import KITS23Pipeline
from .knee_osteoarthritis import KneeOsteoarthritisPipeline
from .lits import LITSPipeline



"""Temporary store paths to images in source directory and their new names."""
import json
import os

from base.extractors.img_id import BaseImgIdExtractor
from base.step import BaseStep


class StoreSourcePaths(BaseStep):
    """Temporary store paths to images in source directory and their new names."""

    def transform(
        self,
        X: list,  # img_paths
    ) -> list:
        """Temporary store paths to images in source directory and their new names.

        Args:
            X (list): List of paths to the images.
        Returns:
            X (list): List of paths to the images.
        """
        source_paths_dict = {self.get_umie_img_path(img_path): img_path for img_path in X}  # {umie_id: src_img_path}

        with open(os.path.join(self.target_path, "source_paths.json"), "w") as temp_file:
            temp_file.write(json.dumps(source_paths_dict))

        return X



"""
This module contains the BaseLabelExtractor class.

BaseLabelExtractor is a base class for constructing a label extractor for an individual dataset.
Label extractors are used to extract labels from the labels file.
Each label extractor at initialization should receive a dictionary of mapping source labels to target labels dict (as seen in labels in dataset config).
Label extractors should implement the _extract method to extract the label from the image path or mask_path.
There is no default implementation of the _extract method, so it must be implemented in the derived class.
"""

from abc import ABC, abstractmethod


class BaseLabelExtractor(ABC):
    """Base class for constructing a label extractor for an individual dataset."""

    def __init__(self, labels: dict):
        """Initialize the BaseLabelExtractor class."""
        self.labels = labels

    def __call__(self, img_path: str, mask_path: str) -> list:
        """
        Extract labels from the labels file.

        Args:
            img_path (str): The path to the image file.

        Returns:
            list: A list of labels extracted from the labels file.
        """
        return self._extract(img_path, mask_path)

    @abstractmethod
    def _extract(self, img_path: str, mask_path: str) -> list:
        """
        Extract the label from the image path.

        Args:
            img_path (str): The path of the image file.

        Returns:
            list: The extracted label.
        """
        raise NotImplementedError("The method _extract must be implemented in the derived class.")



"""Get file paths of all images from a source directory."""

import os
from typing import Optional

from base.step import BaseStep


class GetFilePaths(BaseStep):
    """Get file paths of all images from a source directory."""

    def transform(
        self,
        X: list = [],
    ) -> list:
        """Get file paths of all images from a source directory.

        Args:
            X (list): List of paths to the images.
        Returns:
            X (list): List of paths to the images.
        """
        file_paths = self.get_file_paths(self.source_path)
        return file_paths

    def get_file_paths(self, source_path: str) -> list:
        """Get file paths of all images from a source directory.

        Args:
            source_path (str): Path to the source directory.
        Returns:
            file_paths (list): List of paths to the images.
        """
        file_paths = []
        for root, _, filenames in os.walk(source_path):
            for filename in filenames:
                if filename.startswith("."):
                    continue
                if filename.endswith((".csv", ".json", ".xslx", ".yml", ".txt")):
                    continue
                if filename.startswith("LICENSE"):
                    continue
                else:
                    path = os.path.join(root, filename)
                    # Verify if file is not a mask
                    if self.mask_selector is None or self.mask_selector not in path:
                        file_paths.append(path)
        return file_paths



"""Converts dicom files to png images with appropriate color encoding.""" ""
import glob
import os
import warnings

import cv2
import numpy as np
import pydicom
import pydicom.pixel_data_handlers.util as ddh
from pydicom import dcmread
from tqdm import tqdm

from base.step import BaseStep


class ConvertDcm2Png(BaseStep):
    """Converts dicom files to png images with appropriate color encoding."""

    def transform(self, X: list) -> list:
        """Convert dicom files to png images with appropriate color encoding.

        Args:
            X (list): List of paths to the images.
        Returns:
            X (list): List of paths to the images.
        """
        print("Converting dicom to png...")
        if len(X) == 0:
            raise ValueError("No list of files provided.")
        for img_path in tqdm(X):
            if img_path.endswith(".dcm"):
                self.convert_dcm2png(img_path)

        mask_paths = []
        for root, _, filenames in os.walk(self.masks_path):
            for filename in filenames:
                if filename.startswith("."):
                    continue
                elif filename.endswith(".dcm"):
                    path = os.path.join(root, filename)
                    # Verify if file is not a mask
                    if self.mask_selector is not None and self.mask_selector in path:
                        mask_paths.append(path)
        if mask_paths:
            for mask_path in mask_paths:
                self.convert_dcm2png(mask_path)
        else:
            print("Masks not found.")

        root_path = self.source_path
        new_paths = glob.glob(os.path.join(root_path, "**/*.png"), recursive=True)
        return new_paths

    def convert_dcm2png(self, img_path: str) -> None:
        """Convert dicom files to png images with appropriate color encoding.

        Args:
            img_path (str): Path to the image.
        """
        # iterate over found images
        ds = dcmread(img_path)
        ds = self._convert2little_endian(ds, img_path)
        try:
            output = ddh.apply_modality_lut(ds.pixel_array, ds)
            # if window parameters are provided use it to remove redundant data
            output = self._apply_window(output, ds)
            new_path = img_path.replace(".dcm", ".png")
            cv2.imwrite(new_path, output)

        except Exception as e:
            print(f"Error {e} occured while converting {img_path} {ds.is_little_endian}")
            if self.on_error_remove:
                os.remove(img_path)

    def _convert2little_endian(self, ds: pydicom.dataset.FileDataset, img_path: str) -> pydicom.dataset.FileDataset:
        """Convert dicom image to little endian.

        Args:
            ds (pydicom.dataset.FileDataset): Dicom file.
            img_path (str): Path to the image.
        Returns:
            ds (pydicom.dataset.FileDataset): Dicom file.
        """
        # if image is not little endian implicit convert using gdcmconv
        if ds.is_little_endian is False:
            # convert image to little endian
            os.system(f"gdcmconv -w -X -I -i {img_path} -o {img_path}.converted;")  # noqa: E702,E231
            # Read converted image
            ds = dcmread(f"{img_path}.converted")
            # set property to remove *.converted file at the end
            os.system(f"rm {img_path}.converted")
        return ds

    def _get_window_parameters(self, ds: pydicom.dataset.FileDataset) -> tuple:
        """Get window parameters from dicom file.

        Args:
            ds (pydicom.dataset.FileDataset): Dicom file.
        """
        # Sometimes window center is stored as a list of values, sometimes as a single value
        # If it is a list, we take the first value for simplicity
        if self.window_center is None:
            window_center = (
                int(ds.WindowCenter[0])
                if type(ds.WindowCenter) is pydicom.multival.MultiValue
                else int(ds.WindowCenter)
            )
        else:
            window_center = self.window_center

        if self.window_width is None:
            window_width = (
                int(ds.WindowWidth[0]) if type(ds.WindowWidth) is pydicom.multival.MultiValue else int(ds.WindowWidth)
            )
        else:
            window_width = self.window_width
        return window_center, window_width

    def _apply_window(self, output: np.ndarray, ds: pydicom.dataset.FileDataset) -> np.ndarray:
        """Apply window to the image.

        Args:
            output (np.ndarray): Image to apply window to.
            ds (pydicom.dataset.FileDataset): Dicom file.
        Returns:
            np.ndarray: Image data with applied window.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            try:
                window_center, window_width = self._get_window_parameters(ds)
                output = np.clip(
                    output,
                    window_center - window_width / 2,
                    window_center + window_width / 2,
                )
                min_val = np.min(output)
                min_val = -1000 if min_val < -1000 else min_val
                output = output - min_val
                max_val = np.max(output)
                ratio = max_val / 255 if max_val != 0 else 1  # Avoid division by zero

                output = np.divide(output, ratio).astype(int)
            except RuntimeWarning as e:
                print("Runtime warning caught:", e)
                print("Output max value:", max_val)
                print("Output min value:", min_val)
                # Handle the exception or adjust output as necessary
        return output



"""
This module contains the base extractors for different types of data.

The extractors extract specific information from the data, such as image IDs, labels, phase IDs, and study IDs.
They label extractor is used to extract labels from the data.
The other extractors are used to extract relevant information to form UMIE ID from the data.
The extractors included in this module are:
- BaseImgIdExtractor: Extracts image IDs. (individual img file identifier)
- BaseLabelExtractor: Extracts labels.
- BasePhaseIdExtractor: Extracts phase IDs.
- BaseStudyIdExtractor: Extracts study IDs. Each medical imaging examination contains many imgs, study id identifies all the imgs from the same examination.
"""
from .img_id import BaseImgIdExtractor
from .label import BaseLabelExtractor
from .phase_id import BasePhaseIdExtractor
from .study_id import BaseStudyIdExtractor



"""Change img ids to match the format of the rest of the dataset."""
import glob
import json
import os
import shutil
from typing import Callable, Optional

import jsonlines
import numpy as np
from tqdm import tqdm

from base.extractors.img_id import BaseImgIdExtractor
from base.step import BaseStep


class AddUmieIds(BaseStep):
    """Change img ids to match the format of the rest of the dataset."""

    def transform(
        self,
        X: list,
    ) -> list:
        """Change img ids to match the format of the rest of the dataset.

        Args:
            X (list): List of paths to the images.
        Returns:
            list: List of paths to the images with labels.
        """
        print("Adding new ids to the dataset...")
        if len(X) == 0:
            raise ValueError("No list of files provided.")

        self.new_json: list = []
        X.sort(key=lambda x: self.get_umie_id(x))  # Sort the list of paths based on umie_id
        for img_path in tqdm(X):
            self.add_umie_ids(img_path)

        # Update JSON file
        with jsonlines.open(self.json_path, "w") as writer:
            for obj in self.new_json:
                writer.write(obj)

        root_path = os.path.join(self.target_path, f"{self.dataset_uid}_{self.dataset_name}")
        new_paths = glob.glob(os.path.join(root_path, f"**/{self.image_folder_name}/*.png"), recursive=True)
        return new_paths

    def _update_json(self, umie_path: str, mask_path: str) -> None:

        phase_name, study_id, img_id = self.decode_umie_img_path(umie_path)

        """Update JSON file with the infomration about the images."""
        comparative = ""
        if "PRE" in img_id:
            comparative = "PRE"
        elif "POST" in img_id:
            comparative = "POST"

        if mask_path != "":
            mask_path = self.get_path_without_target_path(mask_path)

        img_info = {}
        img_info = {
            "umie_path": self.get_path_without_target_path(umie_path),
            "dataset_name": self.dataset_name,
            "dataset_uid": self.dataset_uid,
            "phase_name": phase_name,
            "comparative": comparative,
            "study_id": str(study_id),
            "umie_id": os.path.basename(umie_path),
            "mask_path": mask_path,
            "labels": [],
        }

        self.new_json.append(img_info)

    def add_umie_ids(self, img_path: str) -> None:
        """Change img ids to match the format of the rest of the dataset.

        Args:
            img_path (str): Path to the image.
        """
        # Extract relevant information from the source path
        # The logic of the extraction functions depends on the dataset

        if self.segmentation_prefix is not None and self.segmentation_prefix in img_path:
            return None

        umie_path = self.get_umie_img_path(img_path)

        if not self.validate_umie_path(img_path):
            return None

        if not os.path.exists(umie_path):
            if self.target_path in img_path:
                os.rename(img_path, umie_path)
            else:
                shutil.copy2(img_path, umie_path)

        umie_mask_path = ""
        if self.mask_folder_name is not None:
            old_mask_path = img_path.replace(self.image_folder_name, self.mask_folder_name)
            new_mask_path = umie_path.replace(self.image_folder_name, self.mask_folder_name)
            if os.path.exists(new_mask_path):
                umie_mask_path = new_mask_path
            if self.mask_folder_name in img_path:
                if os.path.exists(old_mask_path):
                    os.rename(old_mask_path, new_mask_path)
                    umie_mask_path = new_mask_path

        self._update_json(umie_path, umie_mask_path)



"""
This module contains the definition of the BasePhaseIdExtractor class.

The BasePhaseIdExtractor class is a base class for phase ID extractors.
It provides a default implementation for extracting the phase ID from a given path.
Subclasses can override the `_extract` method to provide custom extraction logic.
Phases are the different stages of the medical imaging examination process, e.g. pre-contrast CT, arterial phase CT, venous phase CT, etc.
For most datasets we use broad categories like CT, MRI, X-ray.
"""


class BasePhaseIdExtractor:
    """
    Base class for phase ID extractors.

    This class provides a default implementation for extracting the phase ID from a given path.
    Subclasses can override the `_extract` method to provide custom extraction logic.

    Attributes:
        None

    Methods:
        __call__(self, path: str) -> str: Invokes the phase ID extraction process.
        _extract(self, path: str) -> str: Default implementation of the phase ID extraction.

    """

    def __init__(self, phases: dict[int, str]):
        """
        Initialize the BasePhaseIdExtractor class.

        Args:
            phases (dict[int, str]): A dictionary mapping phase IDs to their corresponding names.

        Returns:
            None
        """
        self.phases = phases

    def __call__(self, path: str) -> str:
        """
        Invoke the phase ID extraction process.

        Args:
            path (str): The path from which to extract the phase ID.

        Returns:
            str: The extracted phase ID.
        """
        return self._extract(path)

    def _extract(self, path: str) -> str:
        """
        Use default implementation of the phase ID extraction.

        This method extracts the base name of the path as the phase ID.

        Args:
            path (str): The path from which to extract the phase ID.

        Returns:
            str: The extracted phase ID.
        """
        # By default, each dataset has a single phase, so we do not need any extraction logic.
        return "0"



"""Copy masks to a new folder structure."""

import glob
import os
import shutil

from base.step import BaseStep


class CopyMasks(BaseStep):
    """Copy masks to a new folder structure."""

    def transform(
        self,
        X: list,  # img_paths
    ) -> list:
        """Copy masks to a new folder structure.

        Args:
            X (list): List of paths to the images.
        Returns:
            list: List of paths to the images with labels.
        """
        print("Copying masks...")
        if len(X) == 0:
            raise ValueError("No list of files provided.")
        mask_paths = glob.glob(os.path.join(self.masks_path, "**/*.png"), recursive=True)
        for mask_path in mask_paths:
            if (
                os.path.exists(mask_path)
                and self.segmentation_prefix is not None
                and self.segmentation_prefix in mask_path
            ):
                self.copy_masks(mask_path)
        return X

    #
    def copy_masks(self, img_path: str) -> None:
        """Copy PNG masks to a new folder structure.

        Args:
            img_path (str): Path to the image.
        """
        img_id = self.img_id_extractor(img_path)
        study_id = self.study_id_extractor(img_path)
        if self.segmentation_prefix not in img_path:
            return None
        if self.mask_selector is not None and self.mask_selector in img_id:
            img_id = img_id.replace(self.mask_selector, "")
        for phase_id in self.phases.keys():
            if self.multiple_masks_selector and any(
                original_mask_selector in img_path for original_mask_selector in self.multiple_masks_selector.keys()
            ):
                continue
            phase_name = self.phases[phase_id]
            new_file_name = f"{self.dataset_uid}_{phase_id}_{study_id}_{img_id}"
            if "." not in new_file_name:
                new_file_name = new_file_name + ".png"
            new_path = os.path.join(
                self.target_path,
                f"{self.dataset_uid}_{self.dataset_name}",
                phase_name,
                self.mask_folder_name,
                new_file_name,
            )

            if not os.path.exists(new_path):
                shutil.copy2(img_path, new_path)



"""Preprocessing pipeline for Brain with hemorrhage dataset."""
import os
from dataclasses import asdict, dataclass, field
from typing import Any

import cv2

from base.extractors import (
    BaseImgIdExtractor,
    BaseLabelExtractor,
    BasePhaseIdExtractor,
    BaseStudyIdExtractor,
)
from base.pipeline import BasePipeline, PipelineArgs
from config.dataset_config import DatasetArgs, brain_with_intracranial_hemorrhage
from steps import (
    AddLabels,
    AddUmieIds,
    ConvertJpg2Png,
    CopyMasks,
    CreateBlankMasks,
    CreateFileTree,
    DeleteTempFiles,
    DeleteTempPng,
    GetFilePaths,
    MasksToBinaryColors,
    RecolorMasks,
)


class ImgIdExtractor(BaseImgIdExtractor):
    """Extractor for image IDs specific to the Brain with hemorrhage dataset."""

    def _extract(self, img_path: str) -> str:
        """Retrieve image id from path."""
        return os.path.basename(img_path)


class StudyIdExtractor(BaseStudyIdExtractor):
    """Extractor for study IDs specific to the Brain with hemorrhage dataset."""

    def _extract(self, img_path: str) -> str:
        """Extract study id from img path."""
        # Study id is name of parent folder of images in source directory
        return os.path.basename(os.path.dirname(os.path.dirname(img_path)))


class PhaseExtractor(BasePhaseIdExtractor):
    """Extractor for phase specific to the Brain with hemorrhage dataset."""

    def _extract(self, img_path: str) -> str:
        """Extract phase from img path."""
        # Phase is the name of folder 2 levels above image in source directory
        phase_name = os.path.basename(os.path.dirname(img_path))
        lowercase_phases = [x.lower() for x in list(self.phases.values())]
        phase_id = list(self.phases.keys())[lowercase_phases.index(phase_name.lower())]
        return str(phase_id) if phase_id else ""


class LabelExtractor(BaseLabelExtractor):
    """Extractor for labels specific to the Brain with hemorrhage dataset."""

    def __init__(self, labels: dict, masks: dict):
        """Initialize the extractor."""
        super().__init__(labels)
        self.target_colors = [mask["target_color"] for mask in masks.values()]

    def _extract(self, img_path: str, mask_path: str) -> list:
        """Extract label from img path."""
        # If there is a mask associated with the image in a source directory, then the label is 'hemorrhage'
        if os.path.exists(mask_path) and self.target_colors in cv2.imread(mask_path):
            return self.labels["brain_hemorrhage"]
        else:
            return self.labels["normal"]


@dataclass
class BrainWithIntracranialHemorrhagePipeline(BasePipeline):
    """Preprocessing pipeline for Brain with hemorrhage dataset."""

    name: str = "brain_with_intracranial_hemorrhage"  # dataset name used in configs
    steps: tuple = (
        ("get_file_paths", GetFilePaths),
        ("create_file_tree", CreateFileTree),
        ("convert_jpg2png", ConvertJpg2Png),
        ("copy_masks", CopyMasks),
        ("masks_to_binary_colors", MasksToBinaryColors),
        ("recolor_masks", RecolorMasks),
        ("add_new_ids", AddUmieIds),
        ("add_labels", AddLabels),
        # Choose either to create blank masks or delete images without masks
        # Recommended to create blank masks because only about 10% images have masks.
        ("create_blank_masks", CreateBlankMasks),
        ("delete_temp_files", DeleteTempFiles),
        ("delete_temp_png", DeleteTempPng),
    )
    dataset_args: DatasetArgs = field(default_factory=lambda: brain_with_intracranial_hemorrhage)
    pipeline_args: PipelineArgs = field(
        default_factory=lambda: PipelineArgs(
            img_prefix=".",  # prefix of the source image file names
            segmentation_prefix="_HGE_Seg",  # prefix of the source mask file names
            mask_selector="_HGE_Seg",
            img_id_extractor=ImgIdExtractor(),
            study_id_extractor=StudyIdExtractor(),
        )
    )

    def prepare_pipeline(self) -> None:
        """Post initialization actions."""
        # Update args with pipeline_args
        self.args: dict[str, Any] = dict(**self.args, **asdict(self.pipeline_args))
        self.args["phase_id_extractor"] = PhaseExtractor(self.args["phases"])
        self.args["label_extractor"] = LabelExtractor(self.args["labels"], self.args["masks"])



"""Create file tree for dataset."""

import json
import os
from typing import Any

from base.step import BaseStep


class CreateFileTree(BaseStep):
    """Create file tree for dataset."""

    def transform(self, X: list) -> list:
        """Create file tree for dataset.

        Args:
            X (list): List of paths to the images.
        Returns:
            X (list): List of paths to the images.
        """
        self.create_file_tree()
        return X

    def _create_dir(self, *filetree: Any) -> None:
        """Create directory if it does not exist.

        Args:
            *filetree (list): List of folder paths.
        """
        filetree_path = os.path.join(*filetree)
        if not os.path.exists(filetree_path):
            os.makedirs(filetree_path)

    def _create_json(self) -> None:
        """Create JSON file for storing information about target datafiles."""
        # Always create a new file
        with open(self.json_path, "w"):
            pass

    def create_file_tree(self) -> None:
        """Create file tree for dataset."""
        self._create_dir(self.target_path, f"{self.dataset_uid}_{self.dataset_name}")
        self._create_json()

        for phase in self.phases.values():
            self._create_dir(self.target_path, f"{self.dataset_uid}_{self.dataset_name}", phase)

            self._create_dir(
                self.target_path,
                f"{self.dataset_uid}_{self.dataset_name}",
                phase,
                self.image_folder_name,
            )

            if self.masks_path:
                self._create_dir(
                    self.target_path,
                    f"{self.dataset_uid}_{self.dataset_name}",
                    phase,
                    self.mask_folder_name,
                )



"""Delete images without masks."""

import glob
import os

import cv2
import jsonlines
import numpy as np
from tqdm import tqdm

from base.step import BaseStep


class DeleteImgsWithNoAnnotations(BaseStep):
    """Delete images without masks."""

    def transform(self, X: str) -> list:
        """Delete images without annotations.

        Args:
            X (list): List of paths to the images.
        Returns:
            X (list): List of paths to the images.
        """
        root_path = os.path.dirname(os.path.dirname(X[0]))

        self.json_lines = {}
        with jsonlines.open(self.json_path, mode="r") as reader:
            for obj in reader:
                umie_file_name = os.path.basename(obj["umie_path"])
                self.json_lines[umie_file_name] = obj

        if self.mask_folder_name:
            mask_paths = glob.glob(f"{os.path.join(root_path, self.mask_folder_name)}/**/*.png", recursive=True)
            self.mask_names = [os.path.basename(mask) for mask in mask_paths]
        print("Deleting images without annotations...")

        for img_path in tqdm(X):
            self.delete_imgs_with_no_annotations(img_path, root_path)

        # Create new list of paths after the deletion
        root_path = os.path.dirname(X[0])
        new_paths = glob.glob(os.path.join(root_path, "**/*.png"), recursive=True)

        with jsonlines.open(self.json_path, mode="w") as writer:
            remaining_files = set(os.path.basename(path) for path in new_paths)
            for k, obj in self.json_lines.items():
                if k in remaining_files:
                    writer.write(obj)

        return new_paths

    def delete_imgs_with_no_annotations(self, img_path: str, root_path: str) -> None:
        """Delete images without any annotations i.e. labels or masks.

        Args:
            img_path (str): Path to the image.
            root_path (str): Path to the root folder.
        """
        img_name = os.path.basename(img_path)
        no_label = False
        if self.json_lines[img_name]["labels"] == []:
            no_label = True
        if self.masks_path:
            mask_path = os.path.join(root_path, self.mask_folder_name, img_name)
            if img_name not in self.mask_names:
                if no_label:
                    # If there is no mask and no label
                    os.remove(img_path)
            # If there is a mask but it is empty
            elif np.unique(cv2.imread(mask_path)).shape[0] == 1:
                if no_label:
                    # If there is a blank mask and no label
                    os.remove(img_path)
                    if os.path.exists(mask_path):
                        os.remove(mask_path)



"""Recolors masks from default color to the color specified in the config."""

import glob
import os

import cv2
import numpy as np

from base.step import BaseStep


class RecolorMasks(BaseStep):
    """Recolors masks from default color to the color specified in the config."""

    def transform(self, X: list) -> list:
        """Recolors masks from default color to the color specified in the config.

        Args:
            X (list): List of paths to the images.
        Returns:
            X (list): List of paths to the images.
        """
        if len(X) == 0:
            raise ValueError("No list of files provided.")
        # Robust to multiple modalities and lack of masks for some images
        root_path = os.path.join(self.target_path, f"{self.dataset_uid}_{self.dataset_name}")
        mask_paths = glob.glob(os.path.join(root_path, f"**/{self.mask_folder_name}/*.png"), recursive=True)
        print("Recoloring masks...")
        for mask_path in mask_paths:
            if os.path.exists(mask_path):
                self.recolor_masks(mask_path)
        return X

    def recolor_masks(self, mask_path: str) -> None:
        """Recolors masks from default color to the color specified in the config.

        Args:
            mask_path (str): Path to the mask.
        """
        mask = cv2.imread(mask_path)
        # changing pixel values
        for mask_color in self.masks.values():
            np.place(mask, mask == mask_color["source_color"], mask_color["target_color"])
        cv2.imwrite(mask_path, mask)



"""Preprocessing pipeline for KITS23 dataset."""

import json
import os
import re
from cProfile import label
from dataclasses import asdict, dataclass, field
from functools import partial
from typing import Any

import cv2
import numpy as np

from base.extractors import BaseImgIdExtractor, BaseLabelExtractor, BaseStudyIdExtractor
from base.pipeline import BasePipeline, PipelineArgs
from config.dataset_config import DatasetArgs, kits23
from constants import MASK_FOLDER_NAME
from steps import (
    AddLabels,
    AddUmieIds,
    ConvertNii2Png,
    CopyMasks,
    CreateBlankMasks,
    CreateFileTree,
    DeleteImgsWithNoAnnotations,
    DeleteTempPng,
    GetFilePaths,
    RecolorMasks,
)


class ImgIdExtractor(BaseImgIdExtractor):
    """Extractor for image IDs specific to the KITS23 dataset."""

    def _extract(self, img_path: str) -> str:
        """Retrieve image id from path."""
        # Image id is in the source file name after the last underscore
        return os.path.basename(img_path).split("_")[-1]


class StudyIdExtractor(BaseStudyIdExtractor):
    """Extractor for study IDs specific to the KITS23 dataset."""

    def _extract(self, img_path: str) -> str:
        """Get study ID for dataset."""
        # Study id is the folder name of all images in the study
        return os.path.basename((os.path.dirname(img_path))).split("_")[-1]


class LabelExtractor(BaseLabelExtractor):
    """Extractor for labels specific to the KITS23 dataset."""

    def __init__(self, labels: dict, labels_path: str, kidney_findings_colors: list):
        """Initialize label extractor."""
        super().__init__(labels)
        with open(labels_path) as f:
            self.labels_list = json.load(f)
        self.kidney_findings_colors = kidney_findings_colors

    def _extract(self, img_path: str, *args: Any) -> list:
        """Extract label from img path."""
        img_id = os.path.basename(img_path)
        root_path = os.path.dirname(os.path.dirname(img_path))
        mask_path = os.path.join(root_path, MASK_FOLDER_NAME, img_id)
        mask = cv2.imread(mask_path)

        # Check if the mask contains the kidney tumor or cyst
        if np.any(np.isin(self.kidney_findings_colors, np.unique(mask))):
            # Study id is between the second and third underscore in the target image id
            study_id_regex = re.match(r"^(?:[^_]*_){2}([^_]+)", img_id)
            study_id = (
                study_id_regex.group(1) if study_id_regex is not None else None
            )  # Study id is between the second and third underscore
            labels = []
            for case in self.labels_list:
                # Find the case with the matching study id
                if case["case_id"] == f"case_{study_id}":
                    # Remove underscores from the label
                    source_label = case["tumor_histologic_subtype"]
                    # We do not include vague labels
                    if source_label in self.labels.keys():
                        labels = self.labels[source_label]
                    break
            return labels
        return []


@dataclass
class KITS23Pipeline(BasePipeline):
    """Preprocessing pipeline for KITS23 dataset."""

    name: str = "kits23"  # dataset name used in configs
    steps: tuple = (
        ("create_file_tree", CreateFileTree),
        ("get_file_paths", GetFilePaths),
        ("convert_nii2png", ConvertNii2Png),
        ("copy_masks", CopyMasks),
        ("add_umie_ids", AddUmieIds),
        ("recolor_masks", RecolorMasks),
        ("add_labels", AddLabels),
        # Choose either to create blank masks or delete images without masks
        ("create_blank_masks", CreateBlankMasks),
        # ("delete_imgs_with_no_annotations", DeleteImgsWithNoAnnotations),
        ("delete_temp_png", DeleteTempPng),
    )
    dataset_args: DatasetArgs = field(default_factory=lambda: kits23)
    pipeline_args: PipelineArgs = field(
        default_factory=lambda: PipelineArgs(
            zfill=2,
            img_id_extractor=ImgIdExtractor(),  #
            study_id_extractor=StudyIdExtractor(),
            window_center=50,  # Window of abddominal cavity CTs
            window_width=400,
            img_prefix="imaging",  # prefix of the source image file names
            segmentation_prefix="segmentation",  # prefix of the source mask file names
            mask_folder_name=MASK_FOLDER_NAME,
        )
    )

    def prepare_pipeline(self) -> None:
        """Post initialization actions."""
        # Update args with pipeline_args
        self.args: dict[str, Any] = dict(**self.args, **asdict(self.pipeline_args))
        kidney_findings_colors = [
            self.args["masks"]["Neoplasm"]["target_color"],
            self.args["masks"]["RenalCyst"]["target_color"],
        ]
        self.args["label_extractor"] = LabelExtractor(
            self.args["labels"], self.args["labels_path"], kidney_findings_colors
        )



"""Converts masks from xml files to png images with appropriate color encoding."""

import glob
import os
import plistlib
import re

import cv2
import numpy as np
from tqdm import tqdm

from base.step import BaseStep


class CreateMasksFromXML(BaseStep):
    """Converts masks from xml files to png images with appropriate color encoding."""

    def transform(self, X: list) -> list:
        """Convert masks from xml files to png images with appropriate color encoding.

        Args:
            X (list): List of paths to the images.
        Returns:
            X (list): List of paths to the images.
        """
        print("Creating masks from xml files...")
        mask_paths = glob.glob(f"{self.masks_path}/**/*.xml", recursive=True)
        # TODO Print warning if no xml paths available in path / path empty.
        # Make prompt whether to use source_path instead, otherwise skip this step,
        # or whole pipeline. If we skip step, all converted images will be wiped.
        for mask_path in tqdm(mask_paths):
            self.create_masks_from_xml(mask_path)
        return X

    def create_masks_from_xml(
        self,
        mask_path: str,
    ) -> None:
        """Convert masks from xml files to png images with appropriate color encoding.

        Args:
            mask_path (str): Path to the mask.
        """
        with open(mask_path, mode="rb") as xml_file:
            segmentations = plistlib.load(xml_file)["Images"]

        study_id = os.path.basename(mask_path).split(".")[0]

        pattern = r"[-+]?\d*\.\d+|\d+"  # Extract numbers from string
        for segmentation in segmentations:
            img_id = segmentation["ImageIndex"]
            img = np.zeros((512, 512), np.uint8)
            for roi in segmentation["ROIs"]:
                if roi["NumberOfPoints"] > 0:
                    points = []
                    for point in roi["Point_px"]:
                        x, y = re.findall(pattern, point)
                        x, y = float(x), float(y)
                        x, y = int(x), int(y)
                        points.append([x, y])
                    # TODO: add case when there is more than one color
                    color = list(self.masks.values())[0]["target_color"]
                    cv2.fillPoly(img, [np.array(points)], (color))

            for phase_id, phase_name in self.phases.items():
                filename_prefix = f"{self.dataset_uid}_{phase_id}_{study_id}"

                new_path = os.path.join(
                    self.target_path,
                    f"{self.dataset_uid}_{self.dataset_name}",
                    phase_name,
                    self.mask_folder_name,
                    f"{filename_prefix}_{str(img_id).zfill(self.zfill)}.png",
                )
                cv2.imwrite(new_path, img)



"""Preprocessing pipeline for Coronahack dataset."""

import os
from dataclasses import asdict, dataclass, field
from typing import Any

import pandas as pd

from base.extractors import BaseImgIdExtractor, BaseLabelExtractor, BaseStudyIdExtractor
from base.pipeline import BasePipeline, PipelineArgs
from config.dataset_config import DatasetArgs, coronahack
from steps import (
    AddLabels,
    AddUmieIds,
    CreateFileTree,
    DeleteImgsWithNoAnnotations,
    GetFilePaths,
)


class ImgIdExtractor(BaseImgIdExtractor):
    """Extractor for image IDs specific to the Coronahack dataset."""

    def _extract(self, img_path: str) -> str:
        """Retrieve image id from path."""
        return "0.png"


class StudyIdExtractor(BaseStudyIdExtractor):
    """Extractor for study IDs specific to the Coronahack dataset."""

    def __init__(self, labels_path: os.PathLike):
        """Initialize study id extractor."""
        self.metadata = pd.read_csv(labels_path)
        self.metadata.rename(columns={"Unnamed: 0": "id"}, inplace=True)

    def _extract(self, img_path: str) -> str:
        """Extract study id from img path."""
        img_name = os.path.split(img_path)[-1]
        img_row = self.metadata.loc[self.metadata["X_ray_image_name"] == img_name]

        if img_row.empty or img_name.endswith("csv"):
            # File not present in csv, or is csv
            return ""

        return img_row["id"].values[0]


class LabelExtractor(BaseLabelExtractor):
    """Extractor for labels specific to the Coronahack dataset."""

    def __init__(self, labels: dict[str, list[int]], labels_path: os.PathLike):
        """Initialize label extractor."""
        super().__init__(labels)
        self.source_labels = pd.read_csv(labels_path)
        self.source_labels.rename(columns={"Unnamed: 0": "id"}, inplace=True)

    def _extract(self, img_path: os.PathLike, *args: Any) -> list[dict[str, int]]:
        """Extract label from img path."""
        img_name = os.path.split(img_path)[-1]
        study_id = img_name.split("_")[2]
        img_row = self.source_labels.loc[self.source_labels["id"] == int(study_id)]
        label = img_row["Label"].values[0]

        radlex_labels = []
        if label == "Pnemonia":
            if img_row["Label_1_Virus_category"].values[0] == "bacteria":
                label = "PneumoniaBacteria"
            elif img_row["Label_1_Virus_category"].values[0] == "Virus":
                label = "PneumoniaVirus"

        if label in self.labels.keys():
            radlex_labels = self.labels[label]

        return radlex_labels


@dataclass
class CoronaHackPipeline(BasePipeline):
    """Preprocessing pipeline for Coronahack Chest XRay dataset."""

    name: str = "coronahack"  # dataset name used in configs
    steps: tuple = (
        ("create_file_tree", CreateFileTree),
        ("get_file_paths", GetFilePaths),
        # add_new_ids is used here to also add labels
        ("add_new_ids", AddUmieIds),
        ("add_labels", AddLabels),
        ("delete_imgs_with_no_annotations", DeleteImgsWithNoAnnotations),
    )
    dataset_args: DatasetArgs = field(default_factory=lambda: coronahack)
    pipeline_args: PipelineArgs = field(
        default_factory=lambda: PipelineArgs(
            zfill=4,
            img_id_extractor=ImgIdExtractor(),
        )
    )

    def prepare_pipeline(self) -> None:
        """Post initialization actions."""
        # Add dataset specific arguments to the pipeline arguments
        self.args: dict[str, Any] = dict(**self.args, **asdict(self.pipeline_args))
        self.args["label_extractor"] = LabelExtractor(self.args["labels"], self.args["labels_path"])
        self.args["study_id_extractor"] = StudyIdExtractor(self.args["labels_path"])



"""Preprocessing pipeline for Alzheimers dataset."""
import os
import re
from dataclasses import asdict, dataclass, field
from typing import Any

from base.extractors import BaseImgIdExtractor, BaseLabelExtractor, BaseStudyIdExtractor
from base.pipeline import BasePipeline, PipelineArgs
from config.dataset_config import DatasetArgs, alzheimers
from constants import IMG_FOLDER_NAME
from steps import (
    AddLabels,
    AddUmieIds,
    ConvertJpg2Png,
    CreateFileTree,
    DeleteTempFiles,
    DeleteTempPng,
    GetFilePaths,
    StoreSourcePaths,
)


class ImgIdExtractor(BaseImgIdExtractor):
    """Extractor for image IDs specific to the Alzheimer's dataset."""

    def _extract(self, img_path: str) -> str:
        """Retrieve image id from path."""
        img_id = os.path.basename(img_path)
        ext = os.path.splitext(img_id)[1]
        if "train" in img_path:
            img_id = "00"
        else:
            img_id = img_id[:2]
        img_id = img_id + ext
        return img_id


class StudyIdExtractor(BaseStudyIdExtractor):
    """Extractor for study IDs specific to the Alzheimer's dataset."""

    # Filenames in source directory are not unique, so additional id is added to each study_id, based on parent folder
    # name to make them unique across the whole dataset.

    # Ids added to study id based on the name of parent folder name for a 'test' source directory.
    ids_dict_test = {
        "NonDemented": "0",
        "VeryMildDemented": "1",
        "MildDemented": "2",
        "ModerateDemented": "3",
    }
    # Ids added to study id based on the name of parent folder name for a 'train' source directory.
    ids_dict_train = {
        "nonDem": "000",
        "verymildDem": "111",
        "mildDem": "222",
        "moderateDem": "333",
    }

    def _extract(self, img_path: str) -> str:
        """Extract study id from img path."""
        basename = os.path.splitext(os.path.basename(img_path))[0]
        # If brackets exist in filename, then study id is within them, else it is 0.
        if "(" in basename:
            pattern = r"[()]"
            study_id = re.split(pattern, basename)[1]
        else:
            study_id = "0"
        # Add identifier based on folder name to make new file name unique across dataset
        if "train" in img_path:
            for id in self.ids_dict_train.keys():
                basename = basename.replace(id, self.ids_dict_train[id])
            study_id = study_id + basename
        else:
            folder = os.path.basename(os.path.dirname(img_path))
            for id in self.ids_dict_test.keys():
                folder = folder.replace(id, self.ids_dict_test[id])
            study_id = folder + study_id
        return study_id


class LabelExtractor(BaseLabelExtractor):
    """Extractor for labels specific to the Alzheimer's dataset."""

    def _extract(self, img_path: str, *args: Any) -> str:
        """Extract label from img path."""
        source_label = os.path.basename(os.path.dirname(img_path))
        radlex_label = self.labels[source_label]
        return radlex_label


@dataclass
class AlzheimersPipeline(BasePipeline):
    """Preprocessing pipeline for Alzheimers dataset."""

    name: str = "alzheimers"  # dataset name used in configs
    steps: tuple = (
        ("create_file_tree", CreateFileTree),
        ("get_file_paths", GetFilePaths),
        ("store_source_paths", StoreSourcePaths),
        ("convert_jpg2png", ConvertJpg2Png),
        ("add_umie_ids", AddUmieIds),
        ("add_labels", AddLabels),
        ("delete_temp_png", DeleteTempPng),
        ("delete_temp_files", DeleteTempFiles),
    )
    dataset_args: DatasetArgs = field(default_factory=lambda: alzheimers)
    pipeline_args: PipelineArgs = field(
        default_factory=lambda: PipelineArgs(
            image_folder_name=IMG_FOLDER_NAME,
            img_prefix="",
            img_id_extractor=ImgIdExtractor(),
            study_id_extractor=StudyIdExtractor(),
        )
    )

    def prepare_pipeline(self) -> None:
        """Post initialization actions."""
        # Add dataset specific arguments to the pipeline arguments
        self.args: dict[str, Any] = dict(**self.args, **asdict(self.pipeline_args))
        self.args["label_extractor"] = LabelExtractor(self.args["labels"])



"""Preprocessing pipeline for Brain Tumor Detection dataset."""
import os
from dataclasses import asdict, dataclass, field
from typing import Any

from base.extractors import BaseImgIdExtractor, BaseLabelExtractor, BaseStudyIdExtractor
from base.pipeline import BasePipeline, PipelineArgs
from config.dataset_config import DatasetArgs, brain_tumor_detection
from steps import (
    AddLabels,
    AddUmieIds,
    ConvertJpg2Png,
    CreateFileTree,
    DeleteTempFiles,
    DeleteTempPng,
    GetFilePaths,
    StoreSourcePaths,
)


class ImgIdExtractor(BaseImgIdExtractor):
    """Extractor for image IDs specific to the Brain Tumor Detection dataset."""

    def _extract(self, img_path: str) -> str:
        """Retrieve image id from path."""
        return "0.png"


class StudyIdExtractor(BaseStudyIdExtractor):
    """Extractor for study IDs specific to the Brain Tumor Detection dataset."""

    # Dictionary used to replace characters in file names to get numerical study_id
    ids_dict = {" ": "0", "n": "1", "o": "2", "N": "3", "Y": "4"}

    def _extract(self, img_path: str) -> str:
        """Extract study id from img path."""
        img_id = os.path.basename(img_path)
        img_basename = os.path.splitext(img_id)[0]
        # study id based on ids in source dataset, with replaced non-numerical characters
        for id in self.ids_dict.keys():
            img_basename = img_basename.replace(id, self.ids_dict[id])
        return img_basename


class LabelExtractor(BaseLabelExtractor):
    """Extractor for labels specific to the Brain Tumor Detection dataset."""

    def _extract(self, img_path: str, *args: Any) -> list:
        """Extract label from img path."""
        if "Y" in os.path.basename(img_path):
            return self.labels["Y"]
        elif "N" in os.path.basename(img_path) or "n" in os.path.basename(img_path):
            return self.labels["N"]
        else:
            return []


@dataclass
class BrainTumorDetectionPipeline(BasePipeline):
    """Preprocessing pipeline for Brain Tumor Detection dataset."""

    name: str = "brain_tumor_detection"  # dataset name used in configs
    steps: tuple = (
        ("create_file_tree", CreateFileTree),
        ("get_file_paths", GetFilePaths),
        ("get_source_paths", StoreSourcePaths),
        ("convert_jpg2png", ConvertJpg2Png),
        ("add_new_ids", AddUmieIds),
        ("add_labels", AddLabels),
        ("delete_temp_files", DeleteTempFiles),
        ("delete_temp_png", DeleteTempPng),
    )
    dataset_args: DatasetArgs = field(default_factory=lambda: brain_tumor_detection)
    pipeline_args: PipelineArgs = field(
        default_factory=lambda: PipelineArgs(
            image_folder_name="Images",
            img_prefix="",
            img_id_extractor=ImgIdExtractor(),
            study_id_extractor=StudyIdExtractor(),
        )
    )

    def prepare_pipeline(self) -> None:
        """Post initialization actions."""
        # Add dataset specific arguments to the pipeline arguments
        self.args: dict[str, Any] = dict(**self.args, **asdict(self.pipeline_args))
        self.args["label_extractor"] = LabelExtractor(self.args["labels"])



"""Combine and recolor multiple masks into one."""

import glob
import os

import cv2
import numpy as np
import PIL.Image
from tqdm import tqdm

from base.step import BaseStep


class CombineMultipleMasks(BaseStep):
    """Combine and recolor multiple masks into one."""

    def transform(self, X: list) -> list:
        """Recolors masks from default color to the color specified in the config.

        Args:
            X (list): List of paths to the images.
        Returns:
            X (list): List of paths to the images.
        """
        if len(X) == 0:
            raise ValueError("No list of files provided.")

        for root, _, filenames in os.walk(self.source_path):
            for filename in filenames:
                if filename.startswith("."):
                    continue
                else:
                    path = os.path.join(root, filename)
                    # Verify if file is not a mask
                    if self.mask_selector in path:
                        self.combine_multiple_masks(path)

        return X

    def combine_multiple_masks(self, mask_path: str) -> None:
        """Recolors masks from default color to the color specified in the config.

        Args:
            mask_path (str): Path to the mask.
        """
        # changing pixel values
        active_selector = [k for k in self.multiple_masks_selector.keys() if k in mask_path]
        if bool(active_selector):
            new_mask_path = mask_path.replace(active_selector[0], "").replace("__", "_")
            if os.path.exists(new_mask_path):
                return
            mask = cv2.imread(mask_path)
            mask_name = self.multiple_masks_selector[active_selector[0]]
            source_color = self.masks[mask_name]["source_color"]
            target_color = self.mask_encodings[mask_name]["target_color"]
            np.place(mask, mask == source_color, target_color)
            for k, v in self.multiple_masks_selector.items():
                other_mask_path = mask_path.replace(active_selector[0], k)
                # other_mask_path = ""
                other_mask = cv2.imread(other_mask_path)
                source_color = self.masks[v]["source_color"]
                target_color = self.masks[v]["target_color"]
                np.place(other_mask, other_mask == source_color, target_color)
                mask[other_mask > 0] = other_mask[other_mask > 0]

            cv2.imwrite(new_mask_path, mask)



"""Preprocessing pipeline for the Stanford Brain MET dataset."""

import os
from dataclasses import asdict, dataclass, field
from typing import Any

from base.extractors import BasePhaseIdExtractor, BaseStudyIdExtractor
from base.pipeline import BasePipeline, PipelineArgs
from config.dataset_config import DatasetArgs, brain_met_share
from steps import AddUmieIds, CopyMasks, CreateFileTree, GetFilePaths, RecolorMasks


class StudyIdExtractor(BaseStudyIdExtractor):
    """Extractor for study IDs specific to the Brain MET dataset."""

    def _extract(self, img_path: str) -> str:
        """Extract study id from img path."""
        # Study name is the folder two levels above the image
        study_id = os.path.basename(os.path.dirname(os.path.dirname(img_path))).split("_")[-1]
        return study_id


class PhaseIdExtractor(BasePhaseIdExtractor):
    """Extractor for phase IDs specific to the Brain MET dataset."""

    def _extract(self, img_path: str) -> str:
        """Extract phase id from img path."""
        # Phase name is the folder one level above the image
        phase_name = os.path.basename(os.path.dirname(img_path))
        phase_id = [key for key, value in self.phases.items() if value == phase_name][0]
        return str(phase_id)


@dataclass
class BrainMETSharePipeline(BasePipeline):
    """Preprocessing pipeline for the Stanford Brain MET dataset."""

    name: str = "brain_met_share"  # dataset name used in configs
    steps: tuple = (
        ("create_file_tree", CreateFileTree),
        ("get_file_paths", GetFilePaths),
        ("copy_png_masks", CopyMasks),
        ("add_new_ids", AddUmieIds),
        ("recolor_masks", RecolorMasks),
    )
    dataset_args: DatasetArgs = field(
        default_factory=lambda: brain_met_share
    )  # Update default value to use default_factory
    pipeline_args: PipelineArgs = field(
        default_factory=lambda: PipelineArgs(
            zfill=3,
            study_id_extractor=StudyIdExtractor(),
            # Phase name is the folder one level above the image
        )
    )

    def prepare_pipeline(self) -> None:
        """Post initialization actions."""
        # Add dataset specific arguments to the pipeline arguments
        self.args: dict[str, Any] = dict(**self.args, **asdict(self.pipeline_args))
        self.args["phase_id_extractor"] = PhaseIdExtractor(self.args["phases"])



"""Converts tif files to png images."""
import glob
import os
import shutil
from typing import Callable

import numpy as np
import PIL
from tqdm import tqdm

from base.extractors.img_id import BaseImgIdExtractor
from base.step import BaseStep


class ConvertTif2Png(BaseStep):
    """Converts tif files to png images."""

    def transform(
        self,
        X: list,  # img_paths
    ) -> list:
        """Convert tif files to png images with appropriate color encoding.

        Args:
            X (list): List of paths to the images.
        Returns:
            list: List of paths to the images with labels.
        """
        print("Converting tif to png...")
        for img_path in tqdm(X):
            if img_path.endswith(".tif") or img_path.endswith(".tiff"):
                self.convert_tif2png(img_path)

        mask_paths = []
        for root, _, filenames in os.walk(self.masks_path):
            for filename in filenames:
                if filename.startswith("."):
                    continue
                else:
                    path = os.path.join(root, filename)
                    # Verify if file is not a mask
                    if self.mask_selector in path:
                        mask_paths.append(path)
        if mask_paths:
            for mask_path in mask_paths:
                self.convert_tif2png(mask_path)
        else:
            print("Masks not found.")

        new_paths = glob.glob(os.path.join(self.source_path, "**/*.png"), recursive=True)
        return new_paths

    def convert_tif2png(self, img_path: str) -> None:
        """Convert tif files to png images.

        Args:
            img_path (str): Path to the image.
        """
        if self.mask_folder_name not in img_path:

            phase_id = self.phase_id_extractor(img_path)
            if phase_id not in self.phases.keys():
                return None
            phase_name = self.phases[phase_id]
            # Changing .tif to .tiff, so images will be readable for PIL
            if self.img_id_extractor(img_path).endswith(".tif"):
                png_filename = self.img_id_extractor(img_path).replace(".tif", ".tiff")
            else:
                png_filename = self.img_id_extractor(img_path)
            # Copy tiff image to target directory
            tiff_path = os.path.join(
                self.target_path,
                f"{self.dataset_uid}_{self.dataset_name}",
                phase_name,
                self.image_folder_name,
                png_filename,
            )
            shutil.copy2(img_path, tiff_path)

            new_path = img_path.replace(".tif", ".png").replace(".tiff", ".png")
        else:
            new_path = img_path.replace(".tif", ".png").replace(".tiff", ".png")
            tiff_path = img_path
        # Image conversion
        try:
            image = PIL.Image.open(tiff_path)
            invert = True if np.min(np.array(image)) < 0 else False
            image = image.convert("L")
            if invert:
                image = PIL.ImageOps.invert(image)
            image.save(new_path, format="png")
            if self.target_path in tiff_path:
                os.remove(tiff_path)
        except IOError:
            print("Image not found")



"""
This module contains the definition of the BaseImgIdExtractor class.

BaseImgIdExtractor is an abstract base class for extracting image IDs from file paths.
If you want to modify the default behaviour of the image ID extraction, you can create a subclass of BaseImgIdExtractor and override the _extract method.
"""

import os
from abc import ABC, abstractmethod


class BaseImgIdExtractor(ABC):
    """Abstract base class for extracting image IDs from file paths."""

    def __call__(self, path: str) -> str:
        """Extract image id from the path.

        Args:
            path (str): The file path from which to extract the image ID. Usually the source path.

        Returns:
            str: The extracted image ID.

        Examples:
            >>> extractor = BaseImgIdExtractor()
            >>> extractor('/path/to/image_001.png')
            'image_001.png'
        """
        return self._extract(path)

    def _extract(self, path: str) -> str:
        """Use the default method to extract the image ID from the path.

        Args:
            path (str): The file path from which to extract the image ID.

        Returns:
            str: The extracted image ID.
        """
        # by default, extract the image ID from the file name
        return os.path.basename(path)



"""Delete temporary files."""
import os

from base.step import BaseStep


class DeleteTempFiles(BaseStep):
    """Delete temporary files."""

    def transform(
        self,
        X: list,  # img_paths
    ) -> list:
        """Delete temporary files.

        Args:
            X (list): List of paths to the images.
        Returns:
            X (list): List of paths to the images.
        """
        source_paths_file = os.path.join(self.target_path, "source_paths.json")
        if os.path.exists(source_paths_file):
            os.remove(source_paths_file)

        return X



"""Create blank masks for images that don't have masks."""
import glob
import os

import cv2
import jsonlines
import numpy as np
from tqdm import tqdm

from base.step import BaseStep


class CreateBlankMasks(BaseStep):
    """Create blank masks for images that don't have masks."""

    def transform(self, X: list) -> list:
        """Create blank masks for images that don't have masks.

        Args:
            X (list): List of paths to the images.
            target_path (str): Path to the target folder.
        Returns:
            X (list): List of paths to the images.
        """
        mask_paths = glob.glob(
            os.path.join(
                self.target_path, f"**/{self.dataset_uid}_{self.dataset_name}/**/{self.mask_folder_name}/*.png"
            ),
            recursive=True,
        )

        mask_names = [os.path.basename(mask) for mask in mask_paths]
        print("Creating blank masks...")
        self.blank_masks: list = []
        for img_path in tqdm(X):
            img_name = os.path.basename(img_path)
            if img_name not in mask_names:
                self.create_blank_masks(img_path)

        updated_lines = []
        with jsonlines.open(self.json_path, mode="r") as reader:
            for obj in reader:
                if obj["umie_path"] in self.blank_masks:
                    mask_path = self.get_umie_mask_path(obj["umie_path"])
                    obj["mask_path"] = self.get_path_without_target_path(mask_path)

                updated_lines.append(obj)

        with jsonlines.open(self.json_path, mode="w") as writer:
            for obj in updated_lines:
                writer.write(obj)

        return X

    def create_blank_masks(self, img_path: str) -> None:
        """Create blank masks for images that don't have masks.

        Args:
            img_path (str): Path to the image.
        """
        img_name = os.path.basename(img_path)
        img = cv2.imread(img_path)

        new_path = os.path.dirname(os.path.dirname(img_path))
        new_path = os.path.join(new_path, self.mask_folder_name, img_name)
        mask = np.zeros(img.shape, np.uint8)  # Create a black mask
        cv2.imwrite(new_path, mask)

        umie_path = new_path.replace(self.mask_folder_name, self.image_folder_name)
        self.blank_masks.append(self.get_path_without_target_path(umie_path))



"""Constants used in the project."""

IMG_FOLDER_NAME = "Images"  # name of the folder with images in each target dataset
MASK_FOLDER_NAME = "Masks"  # name of the folder with masks in each target dataset

TARGET_PATH = "./data/"  # name of the folder, where processed outputs will be placed



"""
Base pipeline class.

Each source dataset in UMIE has its own pipeline.
Each pipeline is based on the BasePipeline class.
The pipeline is constructed from a set of steps, where each step is a class that processes the data in a specific way.
The pipeline is constructed from the steps and arguments passed to the pipeline.
The pipeline is used to process the dataset and save the processed dataset to the target directory.
"""

from abc import abstractmethod
from dataclasses import asdict, dataclass
from typing import Callable, Optional

from sklearn.pipeline import Pipeline

from base.extractors import (
    BaseImgIdExtractor,
    BasePhaseIdExtractor,
    BaseStudyIdExtractor,
)
from base.step import BaseStep
from config.dataset_config import DatasetArgs
from constants import IMG_FOLDER_NAME, MASK_FOLDER_NAME


@dataclass
class PathArgs:
    """Path arguments. These arguments are user defined in runner_config."""

    source_path: str  # path to the source dataset that is being processed, needs to be filled in by the user
    target_path: str  # path to the directory where the processed dataset will be saved, usually TARGET_PATH
    labels_path: Optional[str] = None  # path to the labels file if it is required
    masks_path: Optional[str] = None  # path to the source masks file if it is required


@dataclass
class PipelineArgs:
    """
    Arguments required by the processing pipeline of a given dataset.

    These arguments are pipeline specific.
    Not all datasets require all arguments. Most arguments are optional or have default values.
    """

    image_folder_name: str = IMG_FOLDER_NAME  # name of folder, where images will be stored
    mask_folder_name: str = MASK_FOLDER_NAME  # name of folder, where masks will be stored
    img_id_extractor: BaseImgIdExtractor = BaseImgIdExtractor()  # function to extract image id from the image path
    study_id_extractor: BaseStudyIdExtractor = (
        BaseStudyIdExtractor()
    )  # function to extract study id from the image path
    phase_id_extractor: BasePhaseIdExtractor = BasePhaseIdExtractor({})  # function to extract phase from the image path
    zfill: Optional[int] = None  # number of digits to pad the **image id** with
    window_center: Optional[int] = None  # value used to process DICOM images
    window_width: Optional[int] = None  # value used to process DICOM images
    label_extractor: Optional[Callable] = None  # function to get label for the individual image
    img_prefix: Optional[
        str
    ] = None  # prefix of the source image file names, used to differentiate between images and masks
    segmentation_prefix: Optional[
        str
    ] = None  # prefix of the source mask file names, used to differentiate between images and masks
    mask_selector: Optional[
        str
    ] = None  # string included only in masks names, used to differentiate between images and masks
    multiple_masks_selector: Optional[
        dict
    ] = None  # dict including mask selector and its meaning for each mask, used when there are multiple masks and each mask has a different selector


@dataclass  # type: ignore[misc]
class BasePipeline:
    """The base class for constructing a pipeline for an individual dataset."""

    path_args: PathArgs  # arguments passed to the pipeline, user defines them at each run of the pipeline
    dataset_args: DatasetArgs
    name: str  # name of the dataset
    pipeline_args: PipelineArgs  # arguments passed to sklearn pipeline required to process a specific dataset # defined in config
    steps: tuple[tuple[str, BaseStep]]

    def __post_init__(self) -> None:
        """Prepare args for the pipeline."""
        self.args: dict = dict(**asdict(self.path_args), **asdict(self.dataset_args))

        # Run prepare_pipeline only if source path exists.
        if self.args["source_path"]:
            self.prepare_pipeline()

    @property
    def pipeline(self) -> Pipeline:
        """Create a pipeline based on the steps and args."""
        args = self.args
        # Create pipeline and pass args to each step
        return Pipeline(steps=[(step[0], step[1](**args)) for step in self.steps])

    @abstractmethod
    def prepare_pipeline(self) -> None:
        """Prepare pipeline. Function is called in post initialization if source path exists."""
        return



"""Preprocessing pipeline for ChestX-ray14 dataset."""
import os
from dataclasses import asdict, dataclass, field
from typing import Any

import pandas as pd

from base.extractors import (
    BaseImgIdExtractor,
    BaseLabelExtractor,
    BaseStudyIdExtractor,
    study_id,
)
from base.pipeline import BasePipeline, PipelineArgs
from config.dataset_config import DatasetArgs, chest_xray14
from steps import (
    AddLabels,
    AddUmieIds,
    CreateFileTree,
    DeleteImgsWithNoAnnotations,
    DeleteTempFiles,
    GetFilePaths,
    StoreSourcePaths,
)


class ImgIdExtractor(BaseImgIdExtractor):
    """Extractor for image IDs specific to the Chest Xray 14 dataset."""

    def _extract(self, img_path: os.PathLike) -> str:
        """Retrieve image id from path."""
        img_name = os.path.split(img_path)[-1]
        # Study id is the second part of the image name before the first underscore
        img_id = img_name.split("_")[1]

        return img_id


class StudyIdExtractor(BaseStudyIdExtractor):
    """Extractor for study IDs specific to the Chest Xray 14 dataset."""

    def _extract(self, img_path: os.PathLike) -> str:
        """Extract study id from img path."""
        img_name = os.path.split(img_path)[-1]
        # Study id is the first part of the image name before the first underscore
        return img_name.split("_")[0]


class LabelExtractor(BaseLabelExtractor):
    """Extractor for labels specific to the Brain Tumor Detection dataset."""

    def __init__(self, labels: dict[str, str], labels_path: os.PathLike):
        """Initialize the extractor."""
        super().__init__(labels)
        self.source_labels = pd.read_csv(labels_path)[["Image Index", "Finding Labels"]]

    def _extract(self, img_path: os.PathLike, *args: Any) -> list:
        """Extract label from img path."""
        img_name = os.path.basename(img_path)
        img_row = self.source_labels.loc[self.source_labels["Image Index"] == img_name]
        labels = [label for label in img_row["Finding Labels"].values[0].split("|")]
        radlex_labels: list[dict] = []
        for label in labels:
            radlex_labels.append(*self.labels[label])

        return radlex_labels


@dataclass
class ChestXray14Pipeline(BasePipeline):
    """Preprocessing pipeline for Chest Xray 14 dataset."""

    name: str = "chest_xray14"  # dataset name used in configs
    steps: tuple = (
        ("create_file_tree", CreateFileTree),
        ("get_file_paths", GetFilePaths),
        ("store_source_paths", StoreSourcePaths),
        ("add_new_ids", AddUmieIds),
        ("add_labels", AddLabels),
        ("delete_imgs_with_no_annotations", DeleteImgsWithNoAnnotations),
        ("delete_temp_files", DeleteTempFiles),
    )
    dataset_args: DatasetArgs = field(default_factory=lambda: chest_xray14)
    pipeline_args: PipelineArgs = field(
        default_factory=lambda: PipelineArgs(
            zfill=4,
            study_id_extractor=StudyIdExtractor(),
            img_id_extractor=ImgIdExtractor(),
        )
    )

    def prepare_pipeline(self) -> None:
        """Post initialization actions."""
        # Add dataset specific arguments to the pipeline arguments
        self.args: dict[str, Any] = dict(**self.args, **asdict(self.pipeline_args))
        self.args["label_extractor"] = LabelExtractor(self.args["labels"], self.args["labels_path"])



"""Converts nii files to png images with appropriate color encoding."""

import glob
import os
from typing import Callable

import cv2
import nibabel as nib
import numpy as np
from tqdm import tqdm

from base.extractors.img_id import BaseImgIdExtractor
from base.step import BaseStep


class ConvertNii2Png(BaseStep):
    """Converts nii files to png images with appropriate color encoding."""

    def transform(
        self,
        X: list,  # img_paths
    ) -> list:
        """Convert nii files to png images with appropriate color encoding.

        Args:
            X (list): List of paths to the images.
        Returns:
            list: List of paths to the images with labels.
        """
        print("Converting nii to png...")
        if len(X) == 0:
            raise ValueError("No list of files provided.")
        for img_path in tqdm(X):
            if img_path.endswith(".nii.gz"):
                if self.segmentation_prefix in img_path or self.img_prefix in img_path:
                    self.convert_nii2png(img_path)
        new_paths = glob.glob(os.path.join(self.source_path, f"**/{self.img_prefix}*.png"), recursive=True)
        return new_paths

    def convert_nii2png(self, img_path: str) -> None:
        """Convert nii files to png images with appropriate color encoding.

        Args:
            img_path (str): Path to the image.
        """
        nii_img = nib.load(img_path)
        nii_data = nii_img.get_fdata()
        slices = nii_data.shape[0]
        for idx in range(slices):
            root_path = os.path.dirname(img_path)
            name = os.path.basename(img_path).split(".")[0] + f"_{str(idx).zfill(self.zfill)}.png"
            new_path = os.path.join(root_path, name)
            img = np.array(nii_data[idx, :, :])
            if self.segmentation_prefix not in new_path:
                img = self._apply_window(img)

            cv2.imwrite(new_path, img)

    def _apply_window(self, pixel_data: np.ndarray) -> np.ndarray:
        """Apply window to the image.

        Args:
            pixel_data (np.ndarray): Image data.
        Returns:
            np.ndarray: Image data with applied window.
        """
        # apply window
        pixel_data = np.clip(
            pixel_data,
            self.window_center - self.window_width / 2,
            self.window_center + self.window_width / 2,
        )
        # convert from hounsfield scale (-1000 to 1000) to png scale (0 to 255)
        min = np.min(pixel_data)
        min = -1000 if min < -1000 else min
        pixel_data = pixel_data - min
        ratio = np.max(pixel_data) / 255
        pixel_data = np.divide(pixel_data, ratio).astype(int)
        return pixel_data



"""Preprocessing pipeline for brain tumor classification dataset."""


import os
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Any

from base.extractors import BaseImgIdExtractor, BaseLabelExtractor, BaseStudyIdExtractor
from base.pipeline import BasePipeline, PipelineArgs
from config.dataset_config import DatasetArgs, brain_tumor_classification
from steps import (
    AddLabels,
    AddUmieIds,
    ConvertJpg2Png,
    CreateFileTree,
    DeleteTempFiles,
    DeleteTempPng,
    GetFilePaths,
    StoreSourcePaths,
)


class ImgIdExtractor(BaseImgIdExtractor):
    """Extractor for image IDs specific to the Brain Tumor Classification dataset."""

    def _extract(self, img_path: str) -> str:
        return "0.png"


class StudyIdExtractor(BaseStudyIdExtractor):
    """Extractor for study IDs specific to the Brain Tumor Classification dataset."""

    def _extract(self, img_path: str) -> str:
        """Extract study id from img path. Img names are not unique.

        To make them unique there is system based on image folder path
        /Training/ -> 0 added /Testing/ -> 1 added then there are 4 folders one for each label.
        This fact is used to assign unique ids
        """
        unique_id_conversion_dict = {
            "glioma_tumor": "00",
            "meningioma_tumor": "01",
            "pituitary_tumor": "10",
            "no_tumor": "11",
        }
        unique_id = ""
        if "Training" in img_path:
            unique_id = "0"
        else:
            unique_id = "1"

        image_folder = os.path.basename(os.path.dirname(img_path))
        unique_id = unique_id + unique_id_conversion_dict[image_folder]

        parent_directory = Path(img_path).parent
        files_in_parent_directory = os.listdir(parent_directory)

        # after conversion to png there are additional png files
        jpg_files = [Path(file).stem for file in files_in_parent_directory if file.lower().endswith(".jpg")]
        # makes sure that we get the same order
        jpg_files.sort()

        img_filename = Path(img_path).name
        return unique_id + str(jpg_files.index(Path(img_filename).stem))


class LabelExtractor(BaseLabelExtractor):
    """Extractor for labels specific to the Brain Tumor Classification dataset."""

    def _extract(self, source_img_path: str, *args: Any) -> list:
        image_folder = os.path.basename(os.path.dirname(source_img_path))
        return self.labels[image_folder]


@dataclass
class BrainTumorClassificationPipeline(BasePipeline):
    """Preprocessing pipeline for Brain Tumor Classification dataset."""

    name: str = "Brain_Tumor_Classification"  # dataset name used in configs
    steps: tuple = (
        ("create_file_tree", CreateFileTree),
        ("get_file_paths", GetFilePaths),
        ("store_source_paths", StoreSourcePaths),
        ("convert_jpg2png", ConvertJpg2Png),
        ("add_new_ids", AddUmieIds),
        ("add_labels", AddLabels),
        ("delete_temp_png", DeleteTempPng),
        ("delete_temp_files", DeleteTempFiles),
    )
    dataset_args: DatasetArgs = field(default_factory=lambda: brain_tumor_classification)
    pipeline_args: PipelineArgs = field(
        default_factory=lambda: PipelineArgs(
            image_folder_name="Images",
            img_id_extractor=ImgIdExtractor(),
            study_id_extractor=StudyIdExtractor(),
            img_prefix="",
        )
    )

    def prepare_pipeline(self) -> None:
        """Post initialization actions."""
        self.pipeline_args.label_extractor = LabelExtractor(self.args["labels"])

        self.args: dict[str, Any] = dict(**self.args, **asdict(self.pipeline_args))



"""Preprocessing pipeline for Covid 19 detection dataset."""

import os
from dataclasses import asdict, dataclass, field
from typing import Any

from base.extractors import BaseImgIdExtractor, BaseLabelExtractor, BaseStudyIdExtractor
from base.pipeline import BasePipeline, PipelineArgs
from config.dataset_config import DatasetArgs, covid19_detection
from constants import IMG_FOLDER_NAME
from steps import (
    AddLabels,
    AddUmieIds,
    ConvertJpg2Png,
    CreateFileTree,
    DeleteTempFiles,
    DeleteTempPng,
    GetFilePaths,
    StoreSourcePaths,
)


class ImgIdExtractor(BaseImgIdExtractor):
    """Extractor for image IDs specific to the Covid 19 detection dataset."""

    def _extract(self, img_path: str) -> str:
        """Retrieve image id from path."""
        return "0.png"


class StudyIdExtractor(BaseStudyIdExtractor):
    """Extractor for study IDs specific to the Covid 19 detection dataset."""

    # Images in the dataset has non-unique ids between classes and folders.
    # Dictionaries below are used to make them unique across whole dataset.

    # Ids added to image names in ValData folder based on their classes
    ids_dict_val = {
        "Normal": "000",
        "BacterialPneumonia": "111",
        "ViralPneumonia": "222",
        "COVID-19": "333",
    }
    # Ids added to image names in NonAugmentedTrain folder based on their classes
    ids_dict_non_aug = {
        "Normal": "444",
        "BacterialPneumonia": "555",
        "ViralPneumonia": "666",
        "COVID-19": "777",
    }

    def _extract(self, img_path: str) -> str:
        """Extract study id from img path."""
        """Get study id with added postfix depending on source location to prevent repeated names."""
        img_id = os.path.basename(img_path)
        img_basename = os.path.splitext(img_id)[0]
        if "ValData" in img_path:
            study_id = img_basename + self.ids_dict_val[os.path.basename(os.path.dirname(img_path))]
        elif "NonAugmentedTrain" in img_path:
            study_id = img_basename + self.ids_dict_non_aug[os.path.basename(os.path.dirname(img_path))]
        else:
            study_id = img_basename
        study_id = study_id.replace("_", "")
        return study_id


class LabelExtractor(BaseLabelExtractor):
    """Extractor for labels specific to the Covid 19 detection dataset."""

    def _extract(self, img_path: str, *args: Any) -> list:
        """Extract label from img path."""
        label = os.path.basename(os.path.dirname(img_path))
        return self.labels[label]


@dataclass
class COVID19DetectionPipeline(BasePipeline):
    """Preprocessing pipeline for Covid 19 detection dataset."""

    name: str = "covid19_detection"  # dataset name used in configs
    steps: tuple = (
        ("create_file_tree", CreateFileTree),
        ("get_file_paths", GetFilePaths),
        ("get_source_paths", StoreSourcePaths),
        ("convert_jpg2png", ConvertJpg2Png),
        ("add_new_ids", AddUmieIds),
        ("add_labels", AddLabels),
        ("delete_temp_files", DeleteTempFiles),
        ("delete_temp_png", DeleteTempPng),
    )
    dataset_args: DatasetArgs = field(default_factory=lambda: covid19_detection)
    pipeline_args: PipelineArgs = field(
        default_factory=lambda: PipelineArgs(
            image_folder_name=IMG_FOLDER_NAME,
            img_prefix="",
            img_id_extractor=ImgIdExtractor(),
            study_id_extractor=StudyIdExtractor(),
        )
    )

    def prepare_pipeline(self) -> None:
        """Post initialization actions."""
        # Add dataset specific arguments to the pipeline arguments
        self.args: dict[str, Any] = dict(**self.args, **asdict(self.pipeline_args))
        self.args["label_extractor"] = LabelExtractor(self.args["labels"])



"""Preprocessing pipeline for the Stanford COCA dataset."""

import os
from dataclasses import asdict, dataclass, field
from typing import Any

from base.extractors import BaseImgIdExtractor, BaseStudyIdExtractor
from base.pipeline import BasePipeline, PipelineArgs
from config.dataset_config import DatasetArgs, coca
from steps import (
    AddUmieIds,
    ConvertDcm2Png,
    CreateFileTree,
    CreateMasksFromXML,
    DeleteImgsWithNoAnnotations,
    DeleteTempPng,
    GetFilePaths,
)


class ImgIdExtractor(BaseImgIdExtractor):
    """Extractor for image IDs specific to the Brain Tumor Progression dataset."""

    def _extract(self, img_path: str) -> str:
        """Retrieve image id from path."""
        # Image id is in the source file name after the last underscore
        return os.path.basename(img_path).split("-")[-1]


class StudyIdExtractor(BaseStudyIdExtractor):
    """Extractor for study IDs specific to the Brain Tumor Progression dataset."""

    def _extract(self, img_path: str) -> str:
        """Extract study id from img path."""
        # Study name is the folder two levels above the image
        return os.path.basename(os.path.dirname(os.path.dirname(img_path)))


@dataclass
class COCAPipeline(BasePipeline):
    """Preprocessing pipeline for the Stanford COCA dataset."""

    name: str = "coca"  # dataset name used in configs
    steps: tuple = (
        ("get_file_paths", GetFilePaths),
        ("create_file_tree", CreateFileTree),
        ("convert_dcm2png", ConvertDcm2Png),
        ("create_masks_from_xml", CreateMasksFromXML),
        ("add_new_ids", AddUmieIds),
        # Choose either to create blank masks or delete images without masks
        # ("create_blank_masks", CreateBlankMasks),
        ("delete_imgs_with_no_annotations", DeleteImgsWithNoAnnotations),
        ("delete_temp_png", DeleteTempPng),
    )
    dataset_args: DatasetArgs = field(default_factory=lambda: coca)
    pipeline_args: PipelineArgs = field(
        default_factory=lambda: PipelineArgs(
            zfill=4,
            # Image id is in the source file name after the last underscore
            img_id_extractor=ImgIdExtractor(),
            # Study name is the folder two levels above the image
            study_id_extractor=StudyIdExtractor(),
        )
    )

    def prepare_pipeline(self) -> None:
        """Post initialization actions."""
        # Add dataset specific arguments to the pipeline arguments
        self.args: dict[str, Any] = dict(**self.args, **asdict(self.pipeline_args))



"""Change img ids to match the format of the rest of the dataset."""
import os
from typing import Callable, Optional

import numpy as np
from sklearn.base import TransformerMixin

from base.extractors.img_id import BaseImgIdExtractor
from config.dataset_config import MaskColor
from constants import IMG_FOLDER_NAME, MASK_FOLDER_NAME


class BaseStep(TransformerMixin):
    """Change img ids to match the format of the rest of the dataset."""

    def __init__(
        self,
        source_path: str,  # path to the dataset that is being processed
        target_path: str,  # path to the directory where the processed dataset will be saved path to the source masks file
        dataset_uid: str,  # unique identifier for the dataset
        dataset_name: str,  # name of the dataset
        phases: dict[
            str, str
        ],  # phase_id used for encoding the phase in img name, phase_name used for naming the folder
        image_folder_name: str = IMG_FOLDER_NAME,  # name of folder, where images will be stored
        mask_folder_name: str = MASK_FOLDER_NAME,  # name of folder, where masks will be stored
        img_id_extractor: BaseImgIdExtractor = BaseImgIdExtractor(),  # function to extract image id from the image path
        study_id_extractor: Callable = lambda x: x,  # function to extract study id from the image path
        phase_id_extractor: Callable = lambda x: "0",  # function to extract phase from the image path
        zfill: Optional[int] = None,  # number of digits to pad the image id with
        window_center: Optional[int] = None,  # value used to process DICOM images
        window_width: Optional[int] = None,  # value used to process DICOM images
        label_extractor: Optional[Callable] = None,  # function to get label for the individual image
        img_prefix: Optional[str] = None,  # prefix of the source image file names
        segmentation_prefix: Optional[str] = None,  # prefix of the source mask file names
        mask_selector: Optional[str] = None,  # string included only in masks names
        multiple_masks_selector: Optional[dict] = None,
        labels: dict[str, list[dict[str, float]]] = {},  # some labels have multiple RadLex codes
        masks: dict[str, MaskColor] = {},
        labels_path: Optional[str] = None,  # path to the labels file
        masks_path: Optional[str] = None,  #
    ):
        """
        Initialize a Step object.

        Args:
            source_path (str): Path to the dataset that is being processed.
            target_path (str): Path to the directory where the processed dataset will be saved.
            labels_path (Optional[str]): Path to the labels file.
            masks_path (Optional[str]): Path to the source masks file.
            dataset_uid (str): Unique identifier for the dataset.
            dataset_name (str): Name of the dataset.
            phases (dict[str, str]): Dictionary containing phase_id used for encoding the phase in img name and phase_name used for naming the folder.
            image_folder_name (str, optional): Name of the folder where images will be stored. Defaults to IMG_FOLDER_NAME.
            mask_folder_name (str, optional): Name of the folder where masks will be stored. Defaults to MASK_FOLDER_NAME.
            img_id_extractor (BaseImgIdExtractor, optional): Function to extract image id from the image path. Defaults to BaseImgIdExtractor().
            study_id_extractor (Callable, optional): Function to extract study id from the image path. Defaults to lambda x: x.
            phase_id_extractor (Callable, optional): Function to extract phase from the image path. Defaults to lambda x: "0".
            zfill (Optional[int], optional): Number of digits to pad the image id with. Defaults to None.
            window_center (Optional[int], optional): Value used to process DICOM images. Defaults to None.
            window_width (Optional[int], optional): Value used to process DICOM images. Defaults to None.
            label_extractor (Optional[Callable], optional): Function to get label for the individual image. Defaults to None.
            img_prefix (Optional[str], optional): Prefix of the source image file names. Defaults to None.
            segmentation_prefix (Optional[str], optional): Prefix of the source mask file names. Defaults to None.
            mask_selector (Optional[str], optional): String included only in masks names. Defaults to None.
            multiple_masks_selector (Optional[dict], optional): Dictionary containing multiple masks selectors. Defaults to None.
            labels (dict[str, list[dict[str, float]]], optional): Dictionary containing labels with multiple RadLex codes. Defaults to {}.
            masks (dict[str, MaskColor], optional): Dictionary containing masks. Defaults to {}.
        """
        self.target_path = target_path
        self.dataset_name = dataset_name
        self.dataset_uid = dataset_uid
        self.phases = phases
        self.image_folder_name = image_folder_name
        self.mask_folder_name = mask_folder_name
        self.img_id_extractor = img_id_extractor
        self.study_id_extractor = study_id_extractor
        self.phase_id_extractor = phase_id_extractor
        self.segmentation_prefix = segmentation_prefix
        self.paths_data = np.array([])
        self.source_path = source_path
        self.labels_path = labels_path
        self.masks_path = masks_path
        self.zfill = zfill
        self.window_center = window_center
        self.window_width = window_width
        self.label_extractor = label_extractor
        self.img_prefix = img_prefix
        self.mask_selector = mask_selector
        self.multiple_masks_selector = multiple_masks_selector
        self.labels = labels
        self.masks = masks
        self.json_path = os.path.join(
            self.target_path,
            f"{self.dataset_uid}_{self.dataset_name}",
            f"{self.dataset_uid}_{self.dataset_name}.jsonl",
        )

    def transform(
        self,
        X: list,  # img_paths
    ) -> list:
        """Change img ids to match the format of the rest of the dataset.

        Args:
            X (list): List of image paths.
        """
        raise NotImplementedError("This method should be implemented in the derived class.")

    def get_umie_id(self, img_path: str) -> str:
        """Create a unique identifier for the image.

        Args:
            img_path (str): Path to the image.

        Returns:
            str: Unique identifier for the image.
        """
        img_id = self.img_id_extractor(img_path)
        ext = os.path.splitext(img_path)[1]
        img_id = img_id.replace(ext, ".png")

        study_id = self.study_id_extractor(img_path)
        phase_id = self.phase_id_extractor(img_path)

        umie_id = f"{self.dataset_uid}_{phase_id}_{study_id}_{img_id}"
        return umie_id

    def validate_umie_path(self, img_path: str) -> bool:
        """Check if umie_path is a valid path (a mechanism for discarding imgs).

        Args:
            img_path (str): Path to the image.

        Returns:
            bool: If true, the path is validate.
        """
        img_id = self.img_id_extractor(img_path)
        study_id = self.study_id_extractor(img_path)
        phase_id = self.phase_id_extractor(img_path)

        if img_id == "" or study_id == "" or phase_id == "":
            return False
        return True

    def get_umie_img_path(self, img_path: str) -> str:
        """Create a unique path for the image.

        Args:
            img_path (str): Path to the image.

        Returns:
            str: Unique path for the image.
        """
        umie_id = self.get_umie_id(img_path)

        phase_id = self.phase_id_extractor(img_path)

        if phase_id not in self.phases.keys():
            raise ValueError(f"Phase id {phase_id} not in the list of phases.")
        phase_name = self.phases[phase_id]

        new_path = os.path.join(
            self.target_path,
            f"{self.dataset_uid}_{self.dataset_name}",
            phase_name,
            self.image_folder_name,
            umie_id,
        )
        return new_path

    def get_umie_mask_path(self, img_path: str) -> str:
        """Create a unique path for the mask.

        Args:
            img_path (str): Path to the image.

        Returns:
            str: Unique path for the mask.
        """
        umie_id = os.path.basename(img_path)
        phase_name = self.decode_umie_img_path(img_path)[0]

        new_path = os.path.join(
            self.target_path,
            f"{self.dataset_uid}_{self.dataset_name}",
            phase_name,
            self.mask_folder_name,
            umie_id,
        )
        return new_path

    def decode_umie_img_path(self, umie_path: str) -> tuple:
        """Decode the unique image path.

        Args:
            umie_path (str): Unique image path.

        Returns:
            tuple: Tuple containing the phase name, study id, and image id.
        """
        umie_id = os.path.basename(umie_path)
        umie_path_elements = umie_id.split("_")
        phase_name = self.phases[umie_path_elements[1]]
        study_id = umie_path_elements[2]
        img_id = umie_path_elements[3]
        return phase_name, study_id, img_id

    def get_path_without_target_path(self, path: str) -> str:
        """Get the path without the target path.

        Args:
            path (str): Path to the image.

        Returns:
            str: Path to the image without the target path.
        """
        return os.path.relpath(path, self.target_path)



"""
This module contains the definition of the BaseStudyIdExtractor class.

The BaseStudyIdExtractor class is a base class for study ID extraction. It provides a default implementation
for extracting the study ID from a given path.
Each medical imaging examination contains many imgs, study id identifies all the imgs from the same examination.
"""

import os


class BaseStudyIdExtractor:
    """
    Base class for study ID extractors.

    This class provides a base implementation for extracting study IDs from file paths.
    Subclasses should override the `_extract` method to provide custom extraction logic.
    """

    def __call__(self, path: str) -> str:
        """
        Call the _extract method to extract the study id from the given path.

        Args:
            path (str): The path from which to extract the study id.

        Returns:
            str: The extracted study id.
        """
        return self._extract(path)

    def _extract(self, path: str) -> str:
        """
        Extract the study ID from the given path.

        Args:
            path (str): The path from which to extract the study ID.

        Returns:
            str: The extracted study ID.
        """
        return path



"""Preprocessing pipeline for Alzheimers dataset."""
import os
from dataclasses import asdict, dataclass, field
from typing import Any

from base.extractors import BaseImgIdExtractor, BaseLabelExtractor, BaseStudyIdExtractor
from base.pipeline import BasePipeline, PipelineArgs
from config.dataset_config import DatasetArgs, knee_osteoarthritis
from constants import IMG_FOLDER_NAME
from steps import (
    AddLabels,
    AddUmieIds,
    CreateFileTree,
    DeleteTempFiles,
    GetFilePaths,
    StoreSourcePaths,
)


class ImgIdExtractor(BaseImgIdExtractor):
    """Extractor for image IDs specific to the Knee Osteoarthritis dataset."""

    def _extract(self, img_path: str) -> str:
        """Extract image id from img path."""
        return "0.png"


class StudyIdExtractor(BaseStudyIdExtractor):
    """Extractor for study IDs specific to the Knee Osteoarthritis dataset."""

    def _extract(self, img_path: str) -> str:
        """Extract study id from img path."""
        # replace letters and delete underscore from filenames
        # letters can't be deleted because they make names unique
        # replace letters and delete underscore from filenames
        # letters can't be deleted because they make names unique
        study_id = os.path.splitext(os.path.basename(img_path))[0].replace("R", "0").replace("L", "1").replace("_", "")
        study_id = study_id + os.path.basename(os.path.dirname(img_path))
        return study_id


class LabelExtractor(BaseLabelExtractor):
    """Extractor for labels specific to the Knee Osteoarthritis dataset."""

    def _extract(self, img_path: str, *args: Any) -> list:
        """Extract label from img path."""
        source_label = os.path.basename(os.path.dirname(img_path))
        return self.labels[source_label]


@dataclass
class KneeOsteoarthritisPipeline(BasePipeline):
    """Preprocessing pipeline for Knee Osteoarthritis dataset."""

    name: str = "Knee_Osteoarthritis"  # dataset name used in configs
    steps: tuple = (
        ("create_file_tree", CreateFileTree),
        ("get_file_paths", GetFilePaths),
        ("get_source_paths", StoreSourcePaths),
        ("add_new_ids", AddUmieIds),
        ("add_new_ids", AddLabels),
        ("delete_temp_files", DeleteTempFiles),
    )
    dataset_args: DatasetArgs = field(default_factory=lambda: knee_osteoarthritis)
    pipeline_args: PipelineArgs = field(
        default_factory=lambda: PipelineArgs(
            image_folder_name=IMG_FOLDER_NAME,
            img_id_extractor=ImgIdExtractor(),
            study_id_extractor=StudyIdExtractor(),
        )
    )

    def prepare_pipeline(self) -> None:
        """Post initialization actions."""
        # Add dataset specific arguments to the pipeline arguments
        self.args: dict[str, Any] = dict(**self.args, **asdict(self.pipeline_args))
        self.args["label_extractor"] = LabelExtractor(self.args["labels"])



"""Delete temporary PNG files created by other steps in the source directory. Use with caution some source data may already be in the PNG format."""

import glob
import os

from tqdm import tqdm

from base.step import BaseStep


class DeleteTempPng(BaseStep):
    """Delete temporary png files created by other steps in the source directory."""

    def transform(
        self,
        X: list,  # img_paths
    ) -> list:
        """Delete temporary PNG files created by other steps in the source directory, including nested directories.

        Args:
            X (list): List of paths to the target images.
        Returns:
            list: List of paths to the target images.
        """
        print("Removing temporary PNG files...")
        png_paths = glob.glob(os.path.join(self.source_path, "**", "*.png"), recursive=True)
        for png_path in tqdm(png_paths):
            self.remove_temp_png(png_path)
        return X

    def remove_temp_png(self, png_path: str) -> None:
        """Remove temporary PNG files created by other steps in the source directory.

        Args:
            png_path (str): Path to the PNG file.
        """
        os.remove(png_path)



"""Add labels to the images and masks based on the labels.json file. The step requires the pipeline to specify the function for mapping the images with annotations."""
import glob
import json
import os
from typing import Callable, Optional

import jsonlines
from tqdm import tqdm

from base.extractors.img_id import BaseImgIdExtractor
from base.step import BaseStep


class AddLabels(BaseStep):
    """Add labels to the images and masks based on the function for mapping the images with annotations specified by the pipeline."""

    def transform(
        self,
        X: list,  # img_paths
    ) -> list:
        """Add labels to the images and masks.

        Args:
            X (list): List of paths to the images.
        Returns:
            list: List of paths to the images with labels.
        """
        print("Adding labels...")
        if len(X) == 0:
            raise ValueError("No list of files provided.")
        if os.path.exists(os.path.join(self.target_path, "source_paths.json")):
            source_paths_dict = json.load(open(os.path.join(self.target_path, "source_paths.json")))
        else:
            source_paths_dict = None

        self.json_updates: dict = {}
        for img_path in tqdm(X):
            self.add_labels(img_path, source_paths_dict)

        updated_lines = []
        with jsonlines.open(self.json_path, mode="r") as reader:
            for obj in reader:
                if obj["umie_path"] in self.json_updates.keys():
                    obj["labels"] = self.json_updates[obj["umie_path"]]
                updated_lines.append(obj)

        with jsonlines.open(self.json_path, mode="w") as writer:
            for obj in updated_lines:
                writer.write(obj)

        root_path = os.path.join(self.target_path, f"{self.dataset_uid}_{self.dataset_name}")
        new_paths = glob.glob(os.path.join(root_path, f"**/{self.image_folder_name}/*.png"), recursive=True)
        return new_paths

    def add_labels(self, img_path: str, source_path_dict: Optional[dict] = None) -> None:
        """Add labels to the image and mask based on the label_extractor function specified by the pipeline.

        Args:
            img_path (str): Path to the image.
            labels_list (list): List of labels.
        """
        mask_path = self.get_umie_mask_path(img_path)
        if source_path_dict:
            labels = self.label_extractor(source_path_dict[img_path], mask_path)
        else:
            labels = self.label_extractor(img_path, mask_path)
        if labels:
            key = self.get_path_without_target_path(img_path)
            self.json_updates[key] = labels



"""Preprocessing pipeline for Liver and liver tumor dataset."""
import os
from dataclasses import asdict, dataclass, field
from functools import partial
from typing import Any

import cv2

from base.extractors import BaseImgIdExtractor, BaseLabelExtractor, BaseStudyIdExtractor
from base.pipeline import BasePipeline, PipelineArgs
from config.dataset_config import DatasetArgs, lits
from constants import IMG_FOLDER_NAME, MASK_FOLDER_NAME
from steps import (
    AddLabels,
    AddUmieIds,
    CombineMultipleMasks,
    CopyMasks,
    CreateFileTree,
    DeleteImgsWithNoAnnotations,
    GetFilePaths,
    RecolorMasks,
)


class ImgIdExtractor(BaseImgIdExtractor):
    """Extractor for image IDs specific to the Liver and liver tumor dataset."""

    def _extract(self, img_path: str) -> str:
        """Retrieve image id from path."""
        basename = os.path.basename(img_path)  # .split(".")[0]
        img_id = basename.rsplit("_", 1)[1]
        return img_id


class StudyIdExtractor(BaseStudyIdExtractor):
    """Extractor for study IDs specific to the Liver and liver tumor dataset."""

    def _extract(self, img_path: str) -> str:
        """Retrieve study id from path."""
        basename = os.path.basename(img_path).split(".")[0]
        study_id = basename.rsplit("_", 1)[0].rsplit("-", 1)[1]
        return study_id


class LabelExtractor(BaseLabelExtractor):
    """Extractor for labels specific to the Liver and liver tumor dataset."""

    def __init__(self, labels: dict, target_color: list):
        """Initialize label extractor."""
        super().__init__(labels)
        self.target_color = target_color

    def _extract(self, img_path: str, mask_path: str, *args: Any) -> list:
        """Get image label based on path."""
        mask = cv2.imread(mask_path)
        if mask is not None and (mask == self.target_color).all(axis=-1).any():
            return self.labels["Neoplasm"]
        else:
            return self.labels["NormalityDescriptor"]


@dataclass
class LITSPipeline(BasePipeline):
    """Preprocessing pipeline for Liver and liver tumor dataset."""

    name: str = "Liver_And_Liver_Tumor"  # dataset name used in configs
    steps: tuple = (
        ("get_file_paths", GetFilePaths),
        ("create_file_tree", CreateFileTree),
        ("combine_multiple_masks", CombineMultipleMasks),
        ("copy_masks", CopyMasks),
        ("recolor_masks", RecolorMasks),
        ("add_new_ids", AddUmieIds),
        ("add_labels", AddLabels),
        # Recommended to delete images without masks, because they contain neither liver nor tumor
        ("delete_imgs_with_no_annotations", DeleteImgsWithNoAnnotations),
    )

    dataset_args: DatasetArgs = field(default_factory=lambda: lits)
    pipeline_args: PipelineArgs = field(
        default_factory=lambda: PipelineArgs(
            img_prefix="volume",  # prefix of the source image file names
            mask_selector="segmentation",
            segmentation_prefix="segmentation",
            multiple_masks_selector={"livermask": "liver", "lesionmask": "liver_tumor"},
            img_id_extractor=ImgIdExtractor(),
            study_id_extractor=StudyIdExtractor(),
        )
    )

    def prepare_pipeline(self) -> None:
        """Post initialization actions."""
        # Update args with dataset_args
        self.args: dict[str, Any] = dict(**self.args, **asdict(self.pipeline_args))
        self.args["label_extractor"] = LabelExtractor(
            self.args["labels"], self.args["masks"]["Neoplasm"]["target_color"]
        )



"""Preprocessing pipeline for the Stanford COCA dataset."""

import os
from dataclasses import asdict, dataclass, field
from typing import Any

from base.extractors import (
    BaseImgIdExtractor,
    BasePhaseIdExtractor,
    BaseStudyIdExtractor,
)
from base.pipeline import BasePipeline, PipelineArgs
from config.dataset_config import DatasetArgs, brain_tumor_progression
from steps import (
    AddUmieIds,
    ConvertDcm2Png,
    CopyMasks,
    CreateBlankMasks,
    CreateFileTree,
    DeleteTempFiles,
    GetFilePaths,
    RecolorMasks,
    StoreSourcePaths,
)


class ImgIdExtractor(BaseImgIdExtractor):
    """Extractor for image IDs specific to the Brain Tumor Progression dataset."""

    def _extract(self, img_path: str) -> str:
        """Retrieve image id from path."""
        return os.path.basename(img_path).split("-")[-1]


class StudyIdExtractor(BaseStudyIdExtractor):
    """Extractor for study IDs specific to the Brain Tumor Progression dataset."""

    def _extract(self, img_path: str) -> str:
        """Extract study id from img path."""
        # Getting study id depends on location of the file.
        # Study_id is retrieved in a different way when image already is moved to target directory with new name.
        return os.path.basename(os.path.dirname(os.path.dirname(img_path)))[-5:]


class PhaseIdExtractor(BasePhaseIdExtractor):
    """Extractor for phase IDs specific to the Brain Tumor Progression dataset."""

    def _extract(self, img_path: str, *args: Any) -> str:
        """Extract phase id from img path."""
        for phase in self.phases.keys():
            if self.phases[phase] in img_path:
                return str(phase)
        return ""


@dataclass
class BrainTumorProgressionPipeline(BasePipeline):
    """Preprocessing pipeline for the Brain Tumor Progression dataset."""

    name: str = field(default="brain_tumor_progression")  # dataset name used in configs
    steps: tuple = (
        ("get_file_paths", GetFilePaths),
        ("create_file_tree", CreateFileTree),
        ("store_source_paths", StoreSourcePaths),
        ("convert_dcm2png", ConvertDcm2Png),
        ("copy_png_masks", CopyMasks),
        ("recolor_masks", RecolorMasks),
        ("add_umie_ids", AddUmieIds),
        # optionally delete images with empty masks
        # ("delete_imgs_with_no_annotations", DeleteImgsWithNoAnnotations),
        # or create blank masks
        ("create_blank_masks", CreateBlankMasks),
        ("delete_temp_files", DeleteTempFiles),
    )
    dataset_args: DatasetArgs = field(default_factory=lambda: brain_tumor_progression)
    pipeline_args: PipelineArgs = field(
        default_factory=lambda: PipelineArgs(
            zfill=4,
            # Image id is in the source file name after the last underscore
            img_id_extractor=ImgIdExtractor(),  # lambda x: os.path.basename(x).split("-")[-1],
            # Study name is the folder two levels above the image
            study_id_extractor=StudyIdExtractor(),
            mask_selector="MaskTumor",
            segmentation_prefix="MaskTumor",
        )
    )

    def prepare_pipeline(self) -> None:
        """Post initialization actions."""
        # self.pipeline_args.study_id_extractor = lambda x: self.study_id_extractor(x)
        # self.pipeline_args.img_id_extractor = lambda x: img_id_extractor(x)
        # Add dataset specific arguments to the pipeline arguments
        self.args: dict[str, Any] = dict(**self.args, **asdict(self.pipeline_args))
        self.args["phase_id_extractor"] = PhaseIdExtractor(self.args["phases"])



"""Recolors masks from shades of gray to 2 colors."""
import glob
import os
from typing import Callable

import cv2
from tqdm import tqdm

from base.step import BaseStep


class MasksToBinaryColors(BaseStep):
    """Recolors masks from shades of gray to 2 colors."""

    def transform(self, X: list) -> list:
        """Recolors masks from default color to the color specified in the config.

        Args:
            X (list): List of paths to the images.
        Returns:
            X (list): List of paths to the images.
        """
        root_path = os.path.join(self.target_path, f"{self.dataset_uid}_{self.dataset_name}")
        mask_paths = glob.glob(os.path.join(root_path, f"**/{self.mask_folder_name}/*.png"), recursive=True)

        for mask_path in tqdm(mask_paths):
            if os.path.exists(mask_path):
                self.scale_to_2_colors(mask_path)
        return X

    def scale_to_2_colors(self, mask_path: str) -> None:
        """Recolors masks from default color to the color specified in the config.

        Args:
            mask_path (str): Path to the mask.
        """
        mask = cv2.imread(mask_path)
        (thresh, blackAndWhiteMask) = cv2.threshold(mask, 127, 255, cv2.THRESH_BINARY)
        # changing pixel values
        cv2.imwrite(mask_path, blackAndWhiteMask)



"""Initialization for the steps module.

This module imports and makes available all the specific processing steps used in various data processing pipelines.
Each step is a class that performs a specific operation, such as converting image formats, copying masks, or adding identifiers.
These steps are used to construct the pipelines defined in the `pipelines` module.
"""


from .add_labels import AddLabels
from .add_umie_ids import AddUmieIds
from .combine_multiple_masks import CombineMultipleMasks
from .convert_dcm2png import ConvertDcm2Png
from .convert_jpg2png import ConvertJpg2Png
from .convert_nii2png import ConvertNii2Png
from .convert_tif2png import ConvertTif2Png
from .copy_masks import CopyMasks
from .create_blank_masks import CreateBlankMasks
from .create_file_tree import CreateFileTree
from .create_masks_from_xml import CreateMasksFromXML
from .delete_imgs_with_no_annotations import DeleteImgsWithNoAnnotations
from .delete_temp_files import DeleteTempFiles
from .delete_temp_png import DeleteTempPng
from .get_file_paths import GetFilePaths
from .masks_to_binary_colors import MasksToBinaryColors
from .recolor_masks import RecolorMasks
from .store_source_paths import StoreSourcePaths



"""Converts jpg and jpeg files to png images."""
import glob
import os
from typing import Callable

from PIL import Image
from tqdm import tqdm

from base.extractors.img_id import BaseImgIdExtractor
from base.step import BaseStep


class ConvertJpg2Png(BaseStep):
    """Converts jpg files to png images."""

    def transform(
        self,
        X: list,  # img_paths
    ) -> list:
        """Convert jpg files to png images with appropriate color encoding.

        Args:
            X (list): List of paths to the images.
        Returns:
            list: List of paths to the images with labels.
        """
        print("Converting jpg to png...")
        paths = glob.glob(os.path.join(self.source_path, "**/*.*"), recursive=True)
        for img_path in tqdm(paths):
            if img_path.endswith(".jpg") or img_path.endswith(".jpeg") or img_path.endswith(".JPG"):
                # Convert each jpg or jpeg file to png
                self.convert_jpg2png(img_path)

        root_path = os.path.join(self.target_path, f"{self.dataset_uid}_{self.dataset_name}")
        mask_paths = glob.glob(
            os.path.join(root_path, f"**/{self.mask_folder_name}/*.jpg"), recursive=True
        ) + glob.glob(os.path.join(root_path, f"**/{self.mask_folder_name}/*.jpeg"), recursive=True)
        if mask_paths:
            for mask_path in tqdm(mask_paths):
                self.convert_jpg2png(mask_path)
        else:
            print("Masks not found.")

        # Get paths to all images after conversion
        new_paths = glob.glob(os.path.join(self.source_path, f"**/*{self.img_prefix}*png"), recursive=True)
        return new_paths

    def convert_jpg2png(self, img_path: str) -> None:
        """Convert jpg to png images.

        Args:
            img_path (str): Path to the image.
        """
        # Path for image after conversion
        new_path = img_path.replace(".jpg", ".png").replace(".jpeg", ".png").replace(".JPG", ".png")
        try:
            # Open image, save with new extension and remove old file
            image = Image.open(img_path)
            image.save(new_path, format="png")
            if self.mask_folder_name and self.mask_folder_name in img_path:
                os.remove(img_path)
        except IOError:
            print("Image not found")
